# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Models to compute equilibrium temperature or max intensity in a conductor."""

from typing import Dict, Any, Optional, Union, Type

from thermohl.power import cigre as cigrep
from thermohl.power import rte as rtep
from thermohl.power import ieee as ieeep
from thermohl.power import olla as ollap

from thermohl.solver.base import Args, Solver
from thermohl.solver.slv1d import Solver1D
from thermohl.solver.slv1t import Solver1T
from thermohl.solver.slv3t import Solver3T
from thermohl.solver.slv3t_legacy import Solver3TL

concreteSolverType = Union[Type[Solver1T], Type[Solver3T], Type[Solver3TL], Type[Solver1D]]


def default_values() -> Dict[str, Any]:
    return Args().__dict__


def _factory(
    dic: Optional[Dict[str, Any]] = None, heateq: str = "1t", model: str = "ieee"
) -> Solver:
    solver: concreteSolverType
    if heateq == "1t":
        solver = Solver1T
    elif heateq == "3t":
        solver = Solver3T
    elif heateq == "3tl":
        solver = Solver3TL
    elif heateq == "1d":
        solver = Solver1D
    else:
        raise ValueError()

    if model == "cigre":
        return solver(
            dic,
            cigrep.JouleHeating,
            cigrep.SolarHeating,
            cigrep.ConvectiveCooling,
            cigrep.RadiativeCooling,
        )
    elif model == "ieee":
        return solver(
            dic,
            ieeep.JouleHeating,
            ieeep.SolarHeating,
            ieeep.ConvectiveCooling,
            ieeep.RadiativeCooling,
        )
    elif model == "olla":
        return solver(
            dic,
            ollap.JouleHeating,
            ollap.SolarHeating,
            ollap.ConvectiveCooling,
            ollap.RadiativeCooling,
        )
    elif model == "rte":
        return solver(
            dic,
            rtep.JouleHeating,
            rtep.SolarHeating,
            rtep.ConvectiveCooling,
            rtep.RadiativeCooling,
        )
    else:
        raise ValueError()


def cigre(dic: Optional[Dict[str, Any]] = None, heateq: str = "1t") -> Solver:
    """Get a Solver using CIGRE models.

    Parameters
    ----------
    dic : dict, optional
        Input values. The default is None.
    heateq : str, optional
        Input heat equation.

    """
    return _factory(dic, heateq=heateq, model="cigre")


def ieee(dic: Optional[Dict[str, Any]] = None, heateq: str = "1t") -> Solver:
    """Get a Solver using IEEE models.

    Parameters
    ----------
    dic : dict, optional
        Input values. The default is None.
    heateq : str, optional
        Input heat equation.

    """
    return _factory(dic, heateq=heateq, model="ieee")


def olla(dic: Optional[Dict[str, Any]] = None, heateq: str = "1t") -> Solver:
    """Get a Solver using RTE-olla models.

    Parameters
    ----------
    dic : dict, optional
        Input values. The default is None.
    heateq : str, optional
        Input heat equation.

    """
    return _factory(dic, heateq=heateq, model="olla")


def rte(dic: Optional[Dict[str, Any]] = None, heateq: str = "1t") -> Solver:
    """Get a Solver using RTE models.

    Parameters
    ----------
    dic : dict, optional
        Input values. The default is None.
    heateq : str, optional
        Input heat equation.

    """
    return _factory(dic, heateq=heateq, model="rte")



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class JouleHeating(PowerTerm):
    """Joule heating term."""

    def __init__(
        self,
        I: floatArrayLike,
        D: floatArrayLike,
        d: floatArrayLike,
        A: floatArrayLike,
        a: floatArrayLike,
        km: floatArrayLike,
        ki: floatArrayLike,
        kl: floatArrayLike,
        kq: floatArrayLike,
        RDC20: floatArrayLike,
        T20: floatArrayLike = 20.0,
        f: floatArrayLike = 50.0,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        I : float or np.ndarray
            Transit intensity.
        D : float or np.ndarray
            External diameter.
        d : float or np.ndarray
            core diameter.
        A : float or np.ndarray
            External (total) section.
        a : float or np.ndarray
            core section.
        km : float or np.ndarray
            Coefficient for magnetic effects.
        ki : float or np.ndarray
            Coefficient for magnetic effects.
        kl : float or np.ndarray
            Linear resistance augmentation with temperature.
        kq : float or np.ndarray
            Quadratic resistance augmentation with temperature.
        RDC20 : float or np.ndarray
            Electric resistance per unit length (DC) at 20°C.
        T20 : float or np.ndarray, optional
            Reference temperature. The default is 20.
        f : float or np.ndarray, optional
            Current frequency (Hz). The default is 50.

        """
        self.I = I
        self.D = D
        self.d = d
        self.kem = self._kem(A, a, km, ki)
        self.kl = kl
        self.kq = kq
        self.RDC20 = RDC20
        self.T20 = T20
        self.f = f

    def _rdc(self, T: floatArrayLike) -> floatArrayLike:
        """
        Compute resistance per unit length for direct current.

        Parameters
        ----------
        T : float or np.ndarray
            Temperature array or value at which to compute the resistance.

        Returns
        -------
        float or np.ndarray
            Resistance per unit length for direct current at the given temperature(s).
        """
        dt = T - self.T20
        return self.RDC20 * (1.0 + self.kl * dt + self.kq * dt**2)

    def _ks(self, rdc: floatArrayLike) -> floatArrayLike:
        """
        Compute skin-effect coefficient.

        This method calculates the skin-effect coefficient based on the given
        resistance (rdc) and the object's attributes. The calculation is an
        approximation as described in the RTE's document.

        Parameters:
        rdc (float or np.ndarray): The resistance value(s) for which the skin-effect
                              coefficient is to be computed.

        Returns:
        floatArrayLike: The computed skin-effect coefficient(s).
        """
        z = (
            8
            * np.pi
            * self.f
            * (self.D - self.d) ** 2
            / ((self.D**2 - self.d**2) * 1.0e07 * rdc)
        )
        a = 7 * z**2 / (315 + 3 * z**2)
        b = 56 / (211 + z**2)
        beta = 1.0 - self.d / self.D
        return 1.0 + a * (1.0 - 0.5 * beta - b * beta**2)

    def _kem(
        self,
        A: floatArrayLike,
        a: floatArrayLike,
        km: floatArrayLike,
        ki: floatArrayLike,
    ) -> floatArrayLike:
        """
        Compute magnetic coefficient.

        Parameters
        ----------
        A : float or np.ndarray
            External (total) section.
        a : float or np.ndarray
            Core section.
        km : float or np.ndarray
            Coefficient for magnetic effects.
        ki : float or np.ndarray
            Coefficient for magnetic effects.

        Returns
        -------
        floatArrayLike
            Computed magnetic coefficient.
        """
        s = (
            np.ones_like(self.I)
            * np.ones_like(A)
            * np.ones_like(a)
            * np.ones_like(km)
            * np.ones_like(ki)
        )
        z = s.shape == ()
        if z:
            s = np.array([1.0])
        I_ = self.I * s
        a_ = a * s
        A_ = A * s
        m = a_ > 0.0
        ki_ = ki * s
        kem = km * s
        kem[m] += ki_[m] * I_[m] / ((A_[m] - a_[m]) * 1.0e06)
        if z:
            kem = kem[0]
        return kem

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        rdc = self._rdc(T)
        ks = self._ks(rdc)
        rac = self.kem * ks * rdc
        return rac * self.I**2



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Callable, Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class ConvectiveCoolingBase(PowerTerm):
    """Convective cooling term."""

    def __init__(
        self,
        alt: floatArrayLike,
        azm: floatArrayLike,
        Ta: floatArrayLike,
        ws: floatArrayLike,
        wa: floatArrayLike,
        D: floatArrayLike,
        rho: Callable[[floatArrayLike, floatArrayLike], floatArrayLike],
        mu: Callable[[floatArrayLike], floatArrayLike],
        lambda_: Callable[[floatArrayLike], floatArrayLike],
        **kwargs: Any,
    ):
        self.alt = alt
        self.Ta = Ta
        self.ws = ws
        self.da = np.arcsin(np.sin(np.deg2rad(np.abs(azm - wa) % 180.0)))
        self.D = D

        self.rho = rho
        self.mu = mu
        self.lambda_ = lambda_

    def _value_forced(
        self,
        Tf: floatArrayLike,
        Td: floatArrayLike,
        vm: floatArrayLike,
    ) -> floatArrayLike:
        """
        Compute forced convective cooling value.

        Parameters
        ----------
        Tf : float or np.ndarray
            Temperature field array.
        Td : float or np.ndarray
            Temperature difference array.
        vm : float or np.ndarray
            Velocity magnitude array.

        Returns
        -------
        float or np.ndarray
            Computed forced convective cooling values.
        """
        Re = self.ws * self.D * vm / self.mu(Tf)
        Kp = (
            1.194
            - np.cos(self.da)
            + 0.194 * np.cos(2.0 * self.da)
            + 0.368 * np.sin(2.0 * self.da)
        )
        return (
            Kp
            * np.maximum(1.01 + 1.35 * Re**0.52, 0.754 * Re**0.6)
            * self.lambda_(Tf)
            * Td
        )

    def _value_natural(
        self,
        Td: floatArrayLike,
        vm: floatArrayLike,
    ) -> floatArrayLike:
        """
        Compute natural convective cooling value.

        Parameters
        ----------
        Td : float or np.ndarray
            Temperature difference array.
        vm : float or np.ndarray
            Velocity magnitude array.

        Returns
        -------
        float or np.ndarray
            Natural convective cooling value array.
        """
        return 3.645 * np.sqrt(vm) * self.D**0.75 * np.sign(Td) * np.abs(Td) ** 1.25

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute convective cooling.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        Tf = 0.5 * (T + self.Ta)
        Td = T - self.Ta
        vm = self.rho(Tf, self.alt)
        return np.maximum(self._value_forced(Tf, Td, vm), self._value_natural(Td, vm))



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import List, Optional, Any

import numpy as np

from thermohl import floatArrayLike, intArrayLike, sun
from thermohl.power.power_term import PowerTerm


class _SRad:
    """Solar radiation calculator."""

    def __init__(self, clean: List[float], indus: List[float]):
        """Initialize the solar radiation calculator.
        Parameters
        ----------
        clean : List[float]
            Coefficients for polynomial function to compute atmospheric turbidity in clean air conditions.
        indus : List[float]
            Coefficients for polynomial function to compute atmospheric turbidity in industrial (polluted) air conditions.
        """
        self.clean = clean
        self.indus = indus

    def catm(
        self, x: floatArrayLike, trb: Optional[floatArrayLike] = 0.0
    ) -> floatArrayLike:
        """Compute coefficient for atmosphere turbidity.
        This method calculates the atmospheric turbidity coefficient using a polynomial
        function of the solar altitude. The coefficients of the polynomial are a weighted
        average of the clean air and industrial air coefficients, with the weights
        determined by the turbidity factor.
        Parameters
        ----------
        x : floatArrayLike
            Solar altitude in degrees.
        trb : floatArrayLike
            Factor representing the atmospheric turbidity (0 for clean air, 1 for industrial air).
        Returns
        -------
        floatArrayLike
            Coefficient for atmospheric turbidity.
        """
        omt = 1.0 - trb
        A = omt * self.clean[6] + trb * self.indus[6]
        B = omt * self.clean[5] + trb * self.indus[5]
        C = omt * self.clean[4] + trb * self.indus[4]
        D = omt * self.clean[3] + trb * self.indus[3]
        E = omt * self.clean[2] + trb * self.indus[2]
        F = omt * self.clean[1] + trb * self.indus[1]
        G = omt * self.clean[0] + trb * self.indus[0]
        return A * x**6 + B * x**5 + C * x**4 + D * x**3 + E * x**2 + F * x**1 + G

    def __call__(
        self,
        lat: floatArrayLike,
        alt: floatArrayLike,
        azm: floatArrayLike,
        trb: floatArrayLike,
        month: intArrayLike,
        day: intArrayLike,
        hour: floatArrayLike,
    ) -> floatArrayLike:
        """Compute solar radiation."""
        sa = sun.solar_altitude(lat, month, day, hour)
        sz = sun.solar_azimuth(lat, month, day, hour)
        th = np.arccos(np.cos(sa) * np.cos(sz - azm))
        K = 1.0 + 1.148e-04 * alt - 1.108e-08 * alt**2
        Q = self.catm(np.rad2deg(sa), trb)
        sr = K * Q * np.sin(th)
        return np.where(sr > 0.0, sr, 0.0)


class SolarHeatingBase(PowerTerm):
    """Solar heating term."""

    def __init__(
        self,
        lat: floatArrayLike,
        alt: floatArrayLike,
        azm: floatArrayLike,
        tb: floatArrayLike,
        month: intArrayLike,
        day: intArrayLike,
        hour: floatArrayLike,
        D: floatArrayLike,
        alpha: floatArrayLike,
        est: _SRad,
        srad: Optional[floatArrayLike] = None,
        **kwargs: Any,
    ):
        self.alpha = alpha
        if srad is None:
            self.srad = est(np.deg2rad(lat), alt, np.deg2rad(azm), tb, month, day, hour)
        else:
            self.srad = np.maximum(srad, 0.0)
        self.D = D

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute solar heating.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        return self.alpha * self.srad * self.D * np.ones_like(T)

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        """Compute solar heating derivative."""
        return np.zeros_like(conductor_temperature)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Tuple, Type, Optional, Dict, Any, Callable

import numpy as np
import pandas as pd

from thermohl import floatArrayLike, floatArray, strListLike, intArray
from thermohl.power import PowerTerm
from thermohl.solver.base import Solver as Solver_, _DEFPARAM as DP, _set_dates, reshape
from thermohl.solver.slv1t import Solver1T
from thermohl.utils import quasi_newton_2d


def _profile_mom(ts: float, tc: float, r: floatArrayLike, re: float) -> floatArrayLike:
    """Analytic temperature profile for steady heat equation in cylinder (mono-mat)."""
    return ts + (tc - ts) * (1.0 - (r / re) ** 2)


def _phi(r: floatArrayLike, ri: floatArrayLike, re: floatArrayLike) -> floatArrayLike:
    """Primitive function used in _profile_bim*** functions."""
    ri2 = ri**2
    return (0.5 * (r**2 - ri2) - ri2 * np.log(r / ri)) / (re**2 - ri2)


def _profile_bim_avg_coeffs(
    ri: floatArrayLike, re: floatArrayLike
) -> tuple[floatArrayLike, floatArrayLike]:
    ri2 = ri**2
    re2 = re**2
    a = 0.5 * (re2 - ri2) ** 2 - re2 * ri2 * (2.0 * np.log(re / ri) - 1.0) - ri**4
    b = 2.0 * re2 * (re2 - ri2) * _phi(re, ri, re)
    return a, b


def _profile_bim_avg(
    ts: floatArrayLike, tc: floatArrayLike, ri: floatArrayLike, re: floatArrayLike
) -> floatArrayLike:
    """Analytical formulation for average temperature in _profile_bim."""
    a, b = _profile_bim_avg_coeffs(ri, re)
    return tc - (a / b) * (tc - ts)


class Solver3T(Solver_):

    def __init__(
        self,
        dic: Optional[dict[str, Any]] = None,
        joule: Type[PowerTerm] = PowerTerm,
        solar: Type[PowerTerm] = PowerTerm,
        convective: Type[PowerTerm] = PowerTerm,
        radiative: Type[PowerTerm] = PowerTerm,
        precipitation: Type[PowerTerm] = PowerTerm,
    ):
        super().__init__(dic, joule, solar, convective, radiative, precipitation)
        self.update()

    def _morgan_coefficients(
        self,
    ) -> Tuple[floatArray, floatArray, floatArray, intArray]:
        """
        Calculate coefficients for heat flux between surface and core in steady state.

        Returns:
        --------
        Tuple[numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[int]]
            - c : numpy.ndarray[float]
                Coefficient array for heat flux.
            - D_ : numpy.ndarray[float]
                Array of core diameters, broadcasted to the shape of `c`.
            - d_ : numpy.ndarray[float]
                Array of surface diameters, broadcasted to the shape of `c`.
            - i : numpy.ndarray[int]
                Indices where surface diameter `d_` is greater than 0.
        """
        c = 0.5 * np.ones((self.args.max_len(),))
        D = self.args.D * np.ones_like(c)
        d = self.args.d * np.ones_like(c)
        i = np.nonzero(d > 0.0)[0]
        c[i] -= (d[i] ** 2 / (D[i] ** 2 - d[i] ** 2)) * np.log(D[i] / d[i])
        return c, D, d, i

    def update(self) -> None:
        """
        Updates the solver's internal state by reinitializing several components
        and recalculating the Morgan coefficients.
        This method performs the following steps:
        1. Extends the arguments to their maximum length.
        2. Reinitializes the `jh`, `sh`, `cc`, `rc`, and `pc` components using the updated arguments.
        3. Recalculates the Morgan coefficients using the updated dimensions.
        4. Compresses the arguments.
        Returns:
            None
        """
        self.args.extend_to_max_len()
        self.jh.__init__(**self.args.__dict__)
        self.sh.__init__(**self.args.__dict__)
        self.cc.__init__(**self.args.__dict__)
        self.rc.__init__(**self.args.__dict__)
        self.pc.__init__(**self.args.__dict__)

        self.mgc = self._morgan_coefficients()

        self.args.compress()

    def average(self, ts: floatArray, tc: floatArray) -> floatArrayLike:
        """
        Compute average temperature given surface and core temperature.

        This formula is based on analytical solution in steady-state mode. For
        single material, the formula reduces itself to an usual mean; for
        bi-material conductors, we have geometrical terms to take into account.

        Parameters:
        ts (numpy.ndarray): Array of surface temperatures.
        tc (numpy.ndarray): Array of core temperatures.

        Returns:
        float or numpy.ndarray: Array of average temperatures.
        """
        ta = 0.5 * (ts + tc)
        _, D, d, ix = self.mgc
        ta[ix] = _profile_bim_avg(ts[ix], tc[ix], 0.5 * d[ix], 0.5 * D[ix])
        return ta

    def joule(self, ts: floatArray, tc: floatArray) -> floatArrayLike:
        """
        Calculate the Joule heating effect.

        Parameters:
        ts (numpy.ndarray): Array of surface temperatures.
        tc (numpy.ndarray): Array of core temperatures.

        Returns:
        float or numpy.ndarray: The calculated Joule heating values.

        Notes:
        - The function computes the average temperature `temperature`.
        - Returns the Joule heating values based on the adjusted temperatures.
        """
        ta = self.average(ts, tc)
        return self.jh.value(ta)

    def balance(self, ts: floatArray, tc: floatArray) -> floatArrayLike:
        """
        Calculate the thermal balance.

        This method computes the thermal balance by summing the joule heating,
        specific heat, and subtracting the contributions from the cooling
        components (convection, radiation, and conduction).

        Parameters:
        ts (numpy.ndarray): Array of surface temperatures.
        tc (numpy.ndarray): Array of core temperatures.

        Returns:
        float or numpy.ndarray: The resulting thermal balance.
        """
        return (
            self.joule(ts, tc)
            + self.sh.value(ts)
            - self.cc.value(ts)
            - self.rc.value(ts)
            - self.pc.value(ts)
        )

    def morgan(self, ts: floatArray, tc: floatArray) -> floatArray:
        """
        Computes the Morgan function for given temperature arrays.

        Parameters:
        ts (numpy.ndarray): Array of surface temperatures.
        tc (numpy.ndarray): Array of core temperatures.

        Returns:
        numpy.ndarray: Resulting array after applying the Morgan function.
        """
        c, _, _, _ = self.mgc
        return (tc - ts) - c * self.joule(ts, tc) / (2.0 * np.pi * self.args.l)

    def steady_temperature(
        self,
        Tsg: Optional[floatArrayLike] = None,
        Tcg: Optional[floatArrayLike] = None,
        tol: float = DP.tol,
        maxiter: int = DP.maxiter,
        return_err: bool = False,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """
        Compute the steady-state temperature distribution.
        Parameters:
        -----------
        Tsg : Optional[float or numpy.ndarray], default=None
            Initial guess for the surface temperature. If None, ambient temperature is used.
        Tcg : Optional[float or numpy.ndarray], default=None
            Initial guess for the core temperature. If None, 1.5 times the absolute value of ambient temperature is used.
        tol : float, default=DP.tol
            Tolerance for the quasi-Newton solver.
        maxiter : int, default=DP.maxiter
            Maximum number of iterations for the quasi-Newton solver.
        return_err : bool, default=False
            If True, the error of the solution is included in the returned DataFrame.
        return_power : bool, default=True
            If True, power-related values are included in the returned DataFrame.
        Returns:
        --------
        pd.DataFrame
            DataFrame containing the steady-state temperatures and optionally the error and power-related values.
        """

        # if no guess provided, use ambient temp
        shape = (self.args.max_len(),)
        Tsg = Tsg if Tsg is not None else 1.0 * self.args.Ta
        Tcg = Tcg if Tcg is not None else 1.5 * np.abs(self.args.Ta)
        Tsg_ = Tsg * np.ones(shape)
        Tcg_ = Tcg * np.ones(shape)

        # solve system
        x, y, cnt, err = quasi_newton_2d(
            f1=self.balance,
            f2=self.morgan,
            x0=Tsg_,
            y0=Tcg_,
            relative_tolerance=tol,
            max_iterations=maxiter,
            delta_x=1.0e-03,
            delta_y=1.0e-03,
        )
        if np.max(err) > tol or cnt == maxiter:
            print(f"rstat_analytic max err is {np.max(err):.3E} in {cnt:d} iterations")

        # format output
        z = self.average(x, y)
        df = pd.DataFrame(
            {Solver_.Names.tsurf: x, Solver_.Names.tavg: z, Solver_.Names.tcore: y}
        )

        if return_err:
            df[Solver_.Names.err] = err

        if return_power:
            df[Solver_.Names.pjle] = self.joule(x, y)
            df[Solver_.Names.psol] = self.sh.value(x)
            df[Solver_.Names.pcnv] = self.cc.value(x)
            df[Solver_.Names.prad] = self.rc.value(x)
            df[Solver_.Names.ppre] = self.pc.value(x)

        return df

    def _morgan_transient(self):
        """Morgan coefficients for transient temperature."""
        c, D, d, ix = self.mgc
        c1 = c / (2.0 * np.pi * self.args.l)
        c2 = 0.5 * np.ones_like(c1)
        a, b = _profile_bim_avg_coeffs(0.5 * d[ix], 0.5 * D[ix])
        c2[ix] = a / b
        return c1, c2

    def _transient_temperature_results(self, time, ts, ta, tc, return_power, n):
        dr = {
            Solver_.Names.time: time,
            Solver_.Names.tsurf: ts,
            Solver_.Names.tavg: ta,
            Solver_.Names.tcore: tc,
        }

        if return_power:
            for power in Solver_.Names.powers():
                dr[power] = np.zeros_like(ts)

            for i in range(len(time)):
                dr[Solver_.Names.pjle][i, :] = self.joule(ts[i, :], tc[i, :])
                dr[Solver_.Names.psol][i, :] = self.sh.value(ts[i, :])
                dr[Solver_.Names.pcnv][i, :] = self.cc.value(ts[i, :])
                dr[Solver_.Names.prad][i, :] = self.rc.value(ts[i, :])
                dr[Solver_.Names.ppre][i, :] = self.pc.value(ts[i, :])

        if n == 1:
            keys = list(dr.keys())
            keys.remove(Solver_.Names.time)
            for k in keys:
                dr[k] = dr[k][:, 0]

        return dr

    def transient_temperature(
        self,
        time: floatArray = np.array([]),
        Ts0: Optional[floatArrayLike] = None,
        Tc0: Optional[floatArrayLike] = None,
        return_power: bool = False,
    ) -> Dict[str, Any]:
        """
        Compute transient-state temperature.

        Parameters
        ----------
        time : numpy.ndarray
            A 1D array with times (in seconds) when the temperature needs to be
            computed. The array must contain increasing values (undefined
            behaviour otherwise).
        Ts0 : float
            Initial surface temperature. If set to None, the ambient temperature from
            internal dict will be used. The default is None.
        Tc0 : float
            Initial core temperature. If set to None, the ambient temperature from
            internal dict will be used. The default is None.
        return_power : bool, optional
            Return power term values. The default is False.

        Returns
        -------
        Dict[str, Any]
            A dictionary with temperature and other results (depending on inputs)
            in the keys.

        """
        # get sizes (n for input dict entries, N for time)
        n = self.args.max_len()
        N = len(time)
        if N < 2:
            raise ValueError()

        # get initial temperature
        Ts0 = Ts0 if Ts0 is not None else self.args.Ta
        Tc0 = Tc0 if Tc0 is not None else 1.0 + Ts0

        # get month, day and hours
        month, day, hour = _set_dates(
            self.args.month, self.args.day, self.args.hour, time, n
        )

        # Two dicts, one (dc) with static quantities (with all elements of size
        # n), the other (de) with time-changing quantities (with all elements of
        # size N*n); uk is a list of keys that are in dc but not in de.
        de = dict(
            month=month,
            day=day,
            hour=hour,
            I=reshape(self.args.I, N, n),
            Ta=reshape(self.args.Ta, N, n),
            wa=reshape(self.args.wa, N, n),
            ws=reshape(self.args.ws, N, n),
            Pa=reshape(self.args.Pa, N, n),
            rh=reshape(self.args.rh, N, n),
            pr=reshape(self.args.pr, N, n),
        )
        del (month, day, hour)

        # shortcuts for time-loop
        c1, c2 = self._morgan_transient()
        imc = 1.0 / (self.args.m * self.args.c)

        # init
        ts = np.zeros((N, n))
        ta = np.zeros((N, n))
        tc = np.zeros((N, n))
        ts[0, :] = Ts0
        tc[0, :] = Tc0
        ta[0, :] = self.average(ts[0, :], tc[0, :])

        # main time loop
        for i in range(1, len(time)):
            for k in de.keys():
                self.args[k] = de[k][i, :]
            self.update()
            bal = self.balance(ts[i - 1, :], tc[i - 1, :])
            ta[i, :] = ta[i - 1, :] + (time[i] - time[i - 1]) * bal * imc
            mrg = c1 * (self.jh.value(ta[i, :]) - bal)
            tc[i, :] = ta[i, :] + c2 * mrg
            ts[i, :] = tc[i, :] - mrg

        return self._transient_temperature_results(time, ts, ta, tc, return_power, n)

    @staticmethod
    def _check_target(target, d, max_len):
        """
        Validates and processes the target temperature input.

        Parameters:
        target (str or list): The target temperature(s) to be validated. It can be:
            - "auto": which sets the target to None.
            - A string: which should be one of Solver_.Names.surf, Solver_.Names.avg, or Solver_.Names.core.
            - A list of strings: where each string should be one of Solver_.Names.surf, Solver_.Names.avg, or Solver_.Names.core.
        max_len (int): The expected length of the target list if target is a list.

        Returns:
        numpy.ndarray: An array of target temperatures if the input is valid.

        Raises:
        ValueError: If the target is a string not in the allowed values, or if the target list length does not match max_len, or if any element in the target list is not in the allowed values.
        """
        # check target
        if target == "auto":
            d_ = d * np.ones(max_len)
            target_ = np.array(
                [
                    Solver_.Names.core if d_[i] > 0.0 else Solver_.Names.avg
                    for i in range(max_len)
                ]
            )
        elif isinstance(target, str):
            if target not in [
                Solver_.Names.surf,
                Solver_.Names.avg,
                Solver_.Names.core,
            ]:
                raise ValueError(
                    f"Target temperature should be in "
                    f"{[Solver_.Names.surf, Solver_.Names.avg, Solver_.Names.core]};"
                    f" got {target} instead."
                )
            else:
                target_ = np.array([target for _ in range(max_len)])
        else:
            if len(target) != max_len:
                raise ValueError()
            for t in target:
                if t not in [
                    Solver_.Names.surf,
                    Solver_.Names.avg,
                    Solver_.Names.core,
                ]:
                    raise ValueError()
            target_ = np.array(target)
        return target_

    def _steady_intensity_header(
        self, T: floatArrayLike, target: strListLike
    ) -> Tuple[np.ndarray, Callable]:
        """Format input for ampacity solver."""

        max_len = self.args.max_len()
        Tmax = T * np.ones(max_len)
        target_ = self._check_target(target, self.args.d, max_len)

        # pre-compute indexes
        c, D, d, ix = self.mgc
        a, b = _profile_bim_avg_coeffs(0.5 * d, 0.5 * D)

        js = np.nonzero(target_ == Solver_.Names.surf)[0]
        ja = np.nonzero(target_ == Solver_.Names.avg)[0]
        jc = np.nonzero(target_ == Solver_.Names.core)[0]
        jx = np.intersect1d(ix, ja)

        # get correct input for quasi-newton solver
        def newtheader(i: floatArray, tg: floatArray) -> Tuple[floatArray, floatArray]:
            self.args.I = i
            self.jh.__init__(**self.args.__dict__)
            ts = np.ones_like(tg) * np.nan
            tc = np.ones_like(tg) * np.nan

            ts[js] = Tmax[js]
            tc[js] = tg[js]

            ts[ja] = tg[ja]
            tc[ja] = 2 * Tmax[ja] - ts[ja]
            tc[jx] = (b[jx] * Tmax[jx] - a[jx] * ts[jx]) / (b[jx] - a[jx])

            tc[jc] = Tmax[jc]
            ts[jc] = tg[jc]

            return ts, tc

        return Tmax, newtheader

    def steady_intensity(
        self,
        T: floatArrayLike = np.array([]),
        target: strListLike = "auto",
        tol: float = DP.tol,
        maxiter: int = DP.maxiter,
        return_err: bool = False,
        return_temp: bool = True,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """
        Compute the steady-state intensity for a given temperature profile.
        Parameters:
        -----------
        T : float or numpy.ndarray, optional
            Initial temperature profile. Default is an empty numpy array.
        target : str or List[str], optional
            Target specification for the solver. Default is "auto".
        tol : float, optional
            Tolerance for the solver. Default is DP.tol.
        maxiter : int, optional
            Maximum number of iterations for the solver. Default is DP.maxiter.
        return_err : bool, optional
            If True, return the error in the output DataFrame. Default is False.
        return_temp : bool, optional
            If True, return the temperature profiles in the output DataFrame. Default is True.
        return_power : bool, optional
            If True, return the power profiles in the output DataFrame. Default is True.
        Returns:
        --------
        pd.DataFrame
            DataFrame containing the steady-state intensity and optionally the error, temperature profiles, and power profiles.
        """

        Tmax, newtheader = self._steady_intensity_header(T, target)

        def balance(i: floatArray, tg: floatArray) -> floatArrayLike:
            ts, tc = newtheader(i, tg)
            return self.balance(ts, tc)

        def morgan(i: floatArray, tg: floatArray) -> floatArray:
            ts, tc = newtheader(i, tg)
            return self.morgan(ts, tc)

        # solve system
        s = Solver1T(
            self.args.__dict__,
            type(self.jh),
            type(self.sh),
            type(self.cc),
            type(self.rc),
            type(self.pc),
        )
        r = s.steady_intensity(Tmax, tol=1.0, maxiter=8, return_power=False)
        x, y, cnt, err = quasi_newton_2d(
            balance,
            morgan,
            r[Solver_.Names.transit].values,
            Tmax,
            relative_tolerance=tol,
            max_iterations=maxiter,
            delta_x=1.0e-03,
            delta_y=1.0e-03,
        )
        if np.max(err) > tol or cnt == maxiter:
            print(f"rstat_analytic max err is {np.max(err):.3E} in {cnt:d} iterations")

        # format output
        df = pd.DataFrame({Solver_.Names.transit: x})

        if return_err:
            df["err"] = err

        if return_temp or return_power:
            ts, tc = newtheader(x, y)
            ta = self.average(ts, tc)

            if return_temp:
                df[Solver_.Names.tsurf] = ts
                df[Solver_.Names.tavg] = ta
                df[Solver_.Names.tcore] = tc

            if return_power:
                df[Solver_.Names.pjle] = self.jh.value(ta)
                df[Solver_.Names.psol] = self.sh.value(ts)
                df[Solver_.Names.pcnv] = self.cc.value(ts)
                df[Solver_.Names.prad] = self.rc.value(ts)
                df[Solver_.Names.ppre] = self.pc.value(ts)

        return df



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Utility functions to compute different sun positions from a point on earth.

These positions usually depend on the point latitude and the time. The sun
position is then used to estimate the solar radiation in CIGRE and IEEE
models.
"""

import numpy as np
from thermohl import floatArrayLike, intArrayLike



def utc2solar_hour(hour, minute=0., second=0., lon=0.):
    """convert utc hour to solar hour adding the longitude contribution

    If more than one input are numpy arrays, they should have the same size.

    Parameters
    ----------
    hour : float or numpy.ndarray
        Hour of the day (solar, must be between 0 and 23).
    minute : float or numpy.ndarray, optional
        Minutes on the clock. The default is 0.
    second : float or numpy.ndarray, optional
        Seconds on the clock. The default is 0.
    lon : float or numpy.ndarray, optional
        Longitude (in rad). The default is 0.

    Returns
    -------
    float or numpy.ndarray
        solar hour

    """
    # add 4 min (1/15 of an hour) for every degree of east longitude
    solar_hour = hour % 24 + minute / 60. + second / 3600. + np.rad2deg(lon) / 15.
    return solar_hour



def hour_angle(
    hour: floatArrayLike, minute: floatArrayLike = 0.0, second: floatArrayLike = 0.0
) -> floatArrayLike:
    """Compute hour angle.

    If more than one input are numpy arrays, they should have the same size.

    Parameters
    ----------
    hour : float or numpy.ndarray
        Hour of the day (solar, must be between 0 and 23).
    minute : float or numpy.ndarray, optional
        Minutes on the clock. The default is 0.
    second : float or numpy.ndarray, optional
        Seconds on the clock. The default is 0.

    Returns
    -------
    float or numpy.ndarray
        Hour angle in radians.

    """
    solar_hour = hour % 24 + minute / 60.0 + second / 3600.0
    return np.radians(15.0 * (solar_hour - 12.0))


_csm = np.array([0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334])


def solar_declination(month: intArrayLike, day: intArrayLike) -> floatArrayLike:
    """Compute solar declination.

    If more than one input are numpy arrays, they should have the same size.

    Parameters
    ----------
    month : int or numpy.ndarray
        Month number (must be between 1 and 12)
    day: int or numpy.ndarray
        Day of the month (must be between 1 and 28, 29, 30 or 31 depending on
        month)
    Returns
    -------
    float or numpy.ndarray
        Solar declination in radians.

    """
    doy = _csm[month - 1] + day
    return np.deg2rad(23.46) * np.sin(2.0 * np.pi * (doy + 284) / 365.0)


def solar_altitude(
    lat: floatArrayLike,
    month: intArrayLike,
    day: intArrayLike,
    hour: floatArrayLike,
    minute: floatArrayLike = 0.0,
    second: floatArrayLike = 0.0,
) -> floatArrayLike:
    """Compute solar altitude.

    If more than one input are numpy arrays, they should have the same size.

    Parameters
    ----------
    lat : float or numpy.ndarray
        latitude in radians.
    month : int or numpy.ndarray
        Month number (must be between 1 and 12)
    day: int or numpy.ndarray
        Day of the month (must be between 1 and 28, 29, 30 or 31 depending on
        month)
    hour : float or numpy.ndarray
        Hour of the day (solar, must be between 0 and 23).
    minute : float or numpy.ndarray, optional
        Minutes on the clock. The default is 0.
    second : float or numpy.ndarray, optional
        Seconds on the clock. The default is 0.

    Returns
    -------
    float or numpy.ndarray
        Solar altitude in radians.

    """
    sd = solar_declination(month, day)
    ha = hour_angle(hour, minute=minute, second=second)
    return np.arcsin(np.cos(lat) * np.cos(sd) * np.cos(ha) + np.sin(lat) * np.sin(sd))


def solar_azimuth(
    lat: floatArrayLike,
    month: intArrayLike,
    day: intArrayLike,
    hour: floatArrayLike,
    minute: floatArrayLike = 0.0,
    second: floatArrayLike = 0.0,
) -> floatArrayLike:
    """Compute solar azimuth.

    If more than one input are numpy arrays, they should have the same size.

    Parameters
    ----------
    lat : float or numpy.ndarray
        latitude in radians.
    month : int or numpy.ndarray
        Month number (must be between 1 and 12)
    day: int or numpy.ndarray
        Day of the month (must be between 1 and 28, 29, 30 or 31 depending on
        month)
    hour : float or numpy.ndarray
        Hour of the day (solar, must be between 0 and 23).
    minute : float or numpy.ndarray, optional
        Minutes on the clock. The default is 0.
    second : float or numpy.ndarray, optional
        Seconds on the clock. The default is 0.

    Returns
    -------
    float or numpy.ndarray
        Solar azimuth in radians.

    """
    sd = solar_declination(month, day)
    ha = hour_angle(hour, minute=minute, second=second)
    Xi = np.sin(ha) / (np.sin(lat) * np.cos(ha) - np.cos(lat) * np.tan(sd))
    C = np.where(
        Xi >= 0.0,
        np.where(ha < 0.0, 0.0, np.pi),
        np.where(ha < 0.0, np.pi, 2.0 * np.pi),
    )
    return C + np.arctan(Xi)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from thermohl import floatArrayLike


class Air:
    """Air quantities."""

    @staticmethod
    def volumic_mass(Tc: floatArrayLike, alt: floatArrayLike = 0.0) -> floatArrayLike:
        r"""Compute air volumic mass.

        If both inputs are numpy arrays, they should have the same size.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius).
        alt : float or numpy.ndarray, optional
            Altitude above sea-level. The default is 0.

        Returns
        -------
        float or numpy.ndarray
             Volumic mass in kg.m\ :sup:`-3`\ .

        """
        return (1.293 - 1.525e-04 * alt + 6.379e-09 * alt**2) / (1.0 + 0.00367 * Tc)

    @staticmethod
    def dynamic_viscosity(Tc: floatArrayLike) -> floatArrayLike:
        r"""Compute air dynamic viscosity.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)

        Returns
        -------
        float or numpy.ndarray
             Dynamic viscosity in kg.m\ :sup:`-1`\ .s\ :sup:`-1`\ .

        """
        return (1.458e-06 * (Tc + 273.0) ** 1.5) / (Tc + 383.4)

    @staticmethod
    def thermal_conductivity(Tc: floatArrayLike) -> floatArrayLike:
        r"""Compute air thermal conductivity.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)

        Returns
        -------
        float or numpy.ndarray
             Thermal conductivity in W.m\ :sup:`-1`\ .K\ :sup:`-1`\ .

        """
        return 2.424e-02 + 7.477e-05 * Tc - 4.407e-09 * Tc**2



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Power terms implementation using RTE's olla project choices.
"""

from .air import Air
from .solar_heating import SolarHeating
from .convective_cooling import ConvectiveCooling
from .joule_heating import JouleHeating
from .radiative_cooling import RadiativeCooling



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

from thermohl import floatArrayLike
from thermohl.power import RadiativeCoolingBase


class RadiativeCooling(RadiativeCoolingBase):
    def __init__(
        self,
        Ta: floatArrayLike,
        D: floatArrayLike,
        epsilon: floatArrayLike,
        sigma: float = 5.67e-08,
        **kwargs: Any,
    ):
        r"""Init with args.

        Parameters
        ----------
        Ta : float or np.ndarray
            Ambient temperature (C).
        D : float or np.ndarray
            External diameter (m).
        epsilon : float or np.ndarray
            Emissivity.
        sigma : float, optional
            Stefan-Boltzmann constant in W.m\ :sup:`-2`\ K\ :sup:`4`\ . The
            default is 5.67E-08.

        Returns
        -------

        """
        super().__init__(
            Ta=Ta, D=D, epsilon=epsilon, sigma=sigma, zerok=273.0, **kwargs
        )



# Copyright 2023 Eurobios Mews Labs
# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0


"""Misc. utility code for thermohl project."""

import os

import numpy as np
import pandas as pd
import yaml


def _dict_completion(
    dat: dict, filename: str, check: bool = True, warning: bool = False
) -> dict:
    """Complete input dict with values from file.

    Read dict stored in filename (yaml format) and for each key in it, add it
    to input dict dat if the key is not already in dat.

    Parameters
    ----------
    dat : dict
        Input dict with parameters for power terms.
    warning : bool, optional
        Print a message if a parameter is missing. The default is False.

    Returns
    -------
    dict
        Completed input dict if some parameters were missing.

    """
    fil = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    dfl = yaml.safe_load(open(fil, "r"))
    for k in dfl.keys():
        if k not in dat.keys() or dat[k] is None:
            dat[k] = dfl[k]
            if warning:
                print("Added key %s from default parameters" % (k,))
        elif (
            not isinstance(dat[k], int)
            and not isinstance(dat[k], float)
            and not isinstance(dat[k], np.ndarray)
            and check
        ):
            raise TypeError(
                "element in input dict (key [%s]) must be int, float or numpy.ndarray"
                % (k,)
            )
    return dat


def add_default_parameters(dat: dict, warning: bool = False) -> dict:
    """Add default parameters if there is missing input.

    Parameters
    ----------
    dat : dict
        Input dict with parameters for power terms.
    warning : bool, optional
        Print a message if a parameter is missing. The default is False.

    Returns
    -------
    dict
        Completed input dict if some parameters were missing.

    """
    fil = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "default_values.yaml"
    )
    return _dict_completion(dat, fil, warning=warning)


def add_default_uncertainties(dat: dict, warning: bool = False) -> dict:
    """Add default uncertainty parameters if there is missing input.

    Parameters
    ----------
    dat : dict
        Input dict with parameters for power terms.
    warning : bool, optional
        Print a message if a parameter is missing. The default is False.

    Returns
    -------
    dict
        Completed input dict if some parameters were missing.

    """
    fil = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "default_uncertainties.yaml"
    )
    return _dict_completion(dat, fil, check=False, warning=warning)


def df2dct(df: pd.DataFrame) -> dict:
    """Convert a pandas.DataFrame to a dictionary.

    Would be an equivalent to df.to_dict(orient='numpy.ndarray') if it existed.

    Parameters
    ----------
    df : pandas.DataFrame

    Returns
    -------
    dict
        DESCRIPTION.
    """
    q = df.to_dict(orient="list")
    for k in q.keys():
        if len(q[k]) > 1:
            q[k] = np.array(q[k])
        else:
            q[k] = q[k][0]
    return q


def bisect_v(
    fun: callable,
    a: float,
    b: float,
    shape: tuple[int, ...],
    tol=1.0e-06,
    maxiter=128,
    print_err=False,
) -> tuple[np.ndarray, np.ndarray]:
    """Bisection method to find a zero of a continuous function [a, b] -> R,
    such that f(a) < 0 < f(b).

    The method is vectorized, in the sense that it can in a single call find
    the zeros of several independent real-valued functions with the same input
    range [a, b].
    For this purpose, the `fun` argument should be a Python function taking as
    input a Numpy array of values in [a, b] and returning a Numpy array
    containing the evaluation of each function for the corresponding input.
    This is most efficient if the outputs values of all these functions are
    computed using vectorized Numpy operations as in the example below.

    Parameters
    ----------
    fun: Python function taking a Numpy array and returning a Numpy array of the same shape.
    a : lower bound of [a, b] interval.
    b : upper bound of [a, b] interval.
    shape : shape of the inputs and ouputs of `fun` and thus of the ouputs of `bisect_v`.
    tol : absolute tolerance.
    maxiter : maximum number of iterations.
    print_err : print or not max error and number of iterations at the end.

    Returns
    -------
    x: Numpy array of the same shape as the inputs and outputs of `fun`
        The zero of each function.
    err: Numpy array of the same shape as the inputs and outputs of `fun`
        The convergence error.

    Example
    -------
    >>> c = np.array([1.0, 4.0, 9.0, 16.0])
    >>> def f(x):
    ...     return x**2 - c
    >>> x0, err = bisect_v(f, a=0.0, b=10.0, shape=(4,), tol=1e-10)
    >>> x0
    array([1., 2., 3., 4.])

    """
    a_ = a * np.ones(shape)
    b_ = b * np.ones(shape)

    err = np.abs(b - a)
    count = 1
    while np.nanmax(err) > tol and count <= maxiter:
        x = 0.5 * (a_ + b_)
        y = fun(x)
        i = y < 0
        a_[i] = x[i]
        b_[~i] = x[~i]
        err = np.abs(b_ - a_)
        count += 1
    x = 0.5 * (a_ + b_)
    x[np.isnan(fun(x))] = np.nan
    if print_err:
        print(f"Bisection max err (abs) : {np.max(err):.2E}; count={count}")
    return x, err


# In agreement with Eurobios, this function has been retrieved from the pyntb library,
# in order to remove the external dependency on this library.
# In this library, this function was initially developed under the name qnewt2d_v
def quasi_newton_2d(
    f1: callable,
    f2: callable,
    x0: np.ndarray,
    y0: np.ndarray,
    relative_tolerance: float = 1.0e-12,
    max_iterations: int = 64,
    delta_x: float = 1.0e-03,
    delta_y: float = 1.0e-03,
) -> tuple[np.ndarray, np.ndarray, int, np.ndarray]:
    """
    Two-dimensional quasi-Newton with arrays.

    Apply a 2D quasi newton on a large number of case at the same time, ie solve
    the system [f1(x, y), f2(x, y)] = [0., 0.] in n cases.

    Derivatives are estimated with a second-order centered estimation (ie f1 and
    f2 are evaluated four times at each iteration).

    All return values are arrays of the same size as inputs x0 and y0.

    Parameters
    ----------
    f1 : first component of a 2d function of two variables
    f2 : second component of a 2d function of two variables
    x0 : first component of the initial guess
    y0 : second component of the initial guess
    relative_tolerance : relative tolerance
    max_iterations : max number of iteration
    delta_x : delta for evaluating derivative regarding first component
    delta_y : delta for evaluating derivative regarding second component

    Returns
    -------
    x: first component of solution
    y: second component of solution
    count: number of iterations when exiting the function
    err: relative error when exiting the function

    """
    x = x0.copy()
    y = y0.copy()

    for count in range(max_iterations):
        # Evaluate functions at current x and y
        f1_value = f1(x, y)
        f2_value = f2(x, y)

        # Compute Jacobian matrix using second-order centered differences
        jacobian_11 = (f1(x + delta_x, y) - f1(x - delta_x, y)) / (2 * delta_x)
        jacobian_12 = (f1(x, y + delta_y) - f1(x, y - delta_y)) / (2 * delta_y)
        jacobian_21 = (f2(x + delta_x, y) - f2(x - delta_x, y)) / (2 * delta_x)
        jacobian_22 = (f2(x, y + delta_y) - f2(x, y - delta_y)) / (2 * delta_y)

        # Compute inverse of the Jacobian determinant
        inv_jacobian_det = 1.0 / (jacobian_11 * jacobian_22 - jacobian_12 * jacobian_21)
        err_abs_x = inv_jacobian_det * (jacobian_22 * f1_value - jacobian_12 * f2_value)
        err_abs_y = inv_jacobian_det * (jacobian_11 * f2_value - jacobian_21 * f1_value)

        x -= err_abs_x
        y -= err_abs_y

        # Check for convergence
        err = max(np.nanmax(np.abs(err_abs_x / x)), np.nanmax(np.abs(err_abs_y / y)))
        if err <= relative_tolerance:
            break

    return x, y, count + 1, np.maximum(np.abs(err_abs_x / x), np.abs(err_abs_y / y))



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from typing import Optional, Any

import numpy as np

from thermohl import floatArrayLike, intArrayLike, sun
from thermohl.power import SolarHeatingBase, _SRad

CLEAN_AIR_COEFFICIENTS = [
    -42.0,
    +63.8,
    -1.922,
    0.03469,
    -3.61e-04,
    +1.943e-06,
    -4.08e-09,
]
POLLUTED_AIR_COEFFICIENTS = [0, 0, 0, 0, 0, 0, 0]

solar_radiation = _SRad(clean=CLEAN_AIR_COEFFICIENTS, indus=POLLUTED_AIR_COEFFICIENTS)


def solar_irradiance(
    lat: floatArrayLike,
    month: intArrayLike,
    day: intArrayLike,
    hour: floatArrayLike,
) -> floatArrayLike:
    """Compute solar radiation.

    Difference with IEEE version are neither turbidity or altitude influence.

    Parameters
    ----------
    lat : floatArrayLike
        Latitude in radians.
    month : intArrayLike
        Month (1-12).
    day : intArrayLike
        Day of the month.
    hour : floatArrayLike
        Hour of the day (0-24).

    Returns
    -------
    floatArrayLike
        Solar radiation value. Negative values are set to zero.
    """
    solar_altitude = sun.solar_altitude(lat, month, day, hour)
    atmospheric_coefficient = solar_radiation.catm(np.rad2deg(solar_altitude))
    return np.where(solar_altitude > 0.0, atmospheric_coefficient, 0.0)


class SolarHeating(SolarHeatingBase):
    def __init__(
        self,
        lat: floatArrayLike,
        azm: floatArrayLike,
        month: intArrayLike,
        day: intArrayLike,
        hour: floatArrayLike,
        D: floatArrayLike,
        alpha: floatArrayLike,
        srad: Optional[floatArrayLike] = None,
        **kwargs: Any,
    ):
        r"""Build with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        lat : float or np.ndarray
            Latitude.
        azm : float or np.ndarray
            Azimuth.
        month : int or np.ndarray
            Month number (must be between 1 and 12).
        day : int or np.ndarray
            Day of the month (must be between 1 and 28, 29, 30 or 31 depending on
            month).
        hour : float or np.ndarray
            Hour of the day (solar, must be between 0 and 23).
        D : float or np.ndarray
            external diameter.
        alpha : np.ndarray
            Solar absorption coefficient.
        srad : xxx
            xxx

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        for k in ["alt", "tb"]:
            if k in kwargs.keys():
                kwargs.pop(k)
        super().__init__(
            lat=lat,
            alt=0.0,
            azm=azm,
            tb=0.0,
            month=month,
            day=day,
            hour=hour,
            D=D,
            alpha=alpha,
            est=solar_radiation,
            srad=srad,
            **kwargs,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Optional, Any

import numpy as np

from thermohl import floatArrayLike, intArrayLike, sun as sun
from thermohl.power import PowerTerm


class SolarHeating(PowerTerm):
    """Solar heating term."""

    @staticmethod
    def _solar_radiation(
        lat: floatArrayLike,
        azm: floatArrayLike,
        albedo: floatArrayLike,
        month: intArrayLike,
        day: intArrayLike,
        hour: floatArrayLike,
    ) -> floatArrayLike:
        """Compute solar radiation."""
        sd = sun.solar_declination(month, day)
        sh = sun.hour_angle(hour)
        sa = sun.solar_altitude(lat, month, day, hour)
        Id = 1280.0 * np.sin(sa) / (0.314 + np.sin(sa))
        gs = np.arcsin(np.cos(sd) * np.sin(sh) / np.cos(sa))
        eta = np.arccos(np.cos(sa) * np.cos(gs - azm))
        A = 0.5 * np.pi * albedo * np.sin(sa) + np.sin(eta)
        x = np.sin(sa)
        C = np.piecewise(x, [x < 0.0, x >= 0.0], [lambda x_: 0.0, lambda x_: x_**1.2])
        B = 0.5 * np.pi * (1 + albedo) * (570.0 - 0.47 * Id) * C
        return np.where(sa > 0.0, A * Id + B, 0.0)

    def __init__(
        self,
        lat: floatArrayLike,
        azm: floatArrayLike,
        al: floatArrayLike,
        month: intArrayLike,
        day: intArrayLike,
        hour: floatArrayLike,
        D: floatArrayLike,
        alpha: floatArrayLike,
        srad: Optional[floatArrayLike] = None,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        lat : float or np.ndarray
            Latitude.
        azm : float or np.ndarray
            Azimuth.
        al : float or np.ndarray
            Albedo.
        month : int or np.ndarray
            Month number (must be between 1 and 12).
        day : int or np.ndarray
            Day of the month (must be between 1 and 28, 29, 30 or 31 depending on
            month).
        hour : float or np.ndarray
            Hour of the day (solar, must be between 0 and 23).
        D : float or np.ndarray
            external diameter.
        alpha : float or np.ndarray
            Solar absorption coefficient.
        srad : xxx
            xxx.


        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        self.alpha = alpha
        if srad is None:
            self.srad = SolarHeating._solar_radiation(
                np.deg2rad(lat), np.deg2rad(azm), al, month, day, hour
            )
        else:
            self.srad = srad
        self.D = D

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute solar heating.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        return self.alpha * self.srad * self.D * np.ones_like(T)

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        """Compute solar heating derivative."""
        return np.zeros_like(conductor_temperature)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl import solver

_nprs = 123456


def _solvers(dic=None):
    return [
        solver._factory(dic=dic, heateq=he, model=m)
        for he in ["3t", "3tl"]
        for m in ["rte", "cigre", "ieee", "olla"]
    ]


def test_balance():
    # NB : this one fails only with 'cigre' powers

    tol = 1.0e-09
    np.random.seed(_nprs)
    N = 9999
    dic = dict(
        lat=np.random.uniform(42.0, 51.0, N),
        alt=np.random.uniform(0.0, 1600.0, N),
        azm=np.random.uniform(0.0, 360.0, N),
        month=np.random.randint(1, 13, N),
        day=np.random.randint(1, 31, N),
        hour=np.random.randint(0, 24, N),
        Ta=np.random.uniform(0.0, 30.0, N),
        ws=np.random.uniform(0.0, 7.0, N),
        wa=np.random.uniform(0.0, 90.0, N),
        I=np.random.uniform(40.0, 4000.0, N),
        d=np.random.randint(2, size=N) * solver.default_values()["d"],
    )

    for s in _solvers(dic):
        # compute guess with 1t solver
        s1 = solver._factory(dic=dic, heateq="1t", model="ieee")
        t1 = s1.steady_temperature(
            tol=2.0, maxiter=16, return_err=False, return_power=False
        )
        t1 = t1["t"].values
        # 3t solve
        df = s.steady_temperature(
            Tsg=t1, Tcg=t1, return_err=True, return_power=True, tol=tol, maxiter=64
        )
        # checks
        assert np.all(df["err"] < tol)
        assert np.allclose(
            s.balance(ts=df["t_surf"], tc=df["t_core"]).values, 0.0, atol=tol
        )
        assert np.allclose(
            s.morgan(ts=df["t_surf"], tc=df["t_core"]).values, 0.0, atol=tol
        )


def test_consistency():
    tol = 1.0e-09
    np.random.seed(_nprs)
    N = 9999
    dic = dict(
        lat=np.random.uniform(42.0, 51.0, N),
        alt=np.random.uniform(0.0, 1600.0, N),
        azm=np.random.uniform(0.0, 360.0, N),
        month=np.random.randint(1, 13, N),
        day=np.random.randint(1, 31, N),
        hour=np.random.randint(0, 24, N),
        Ta=np.random.uniform(0.0, 30.0, N),
        ws=np.random.uniform(0.0, 7.0, N),
        wa=np.random.uniform(0.0, 90.0, N),
        d=np.random.randint(2, size=N) * solver.default_values()["d"],
    )

    for s in _solvers(dic):
        for t in ["surf", "avg", "core"]:
            # solve intensity with different targets
            df = s.steady_intensity(
                T=100.0,
                target=t,
                return_err=True,
                return_power=True,
                tol=1.0e-09,
                maxiter=64,
            )
            assert np.all(df["err"] < tol)
            # set args intensity to newly founds ampacities
            s.args.I = df["I"].values
            s.update()
            assert np.allclose(
                s.balance(ts=df["t_surf"], tc=df["t_core"]).values, 0.0, atol=tol
            )
            assert np.allclose(
                s.morgan(ts=df["t_surf"], tc=df["t_core"]).values, 0.0, atol=tol
            )
            # 3t solve
            dg = s.steady_temperature(
                Tsg=df["t_surf"].round(1).values,
                Tcg=df["t_core"].round(1).values,
                return_err=True,
                return_power=True,
                tol=1.0e-09,
                maxiter=64,
            )
            # check consistency
            assert np.allclose(dg["t_" + t].values, 100.0)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl.power import ieee
from thermohl import solver


def test_compare_powers():
    """Compare computed values to hard-coded ones from ieee std 38-2012."""
    dic = solver.default_values()

    # there are a lot of rounding in the standard guide, hence the relatively
    # large tolerances used in our tests ...

    dic["ws"] = 0.61
    dic["wa"] = 0.0
    dic["epsilon"] = 0.8
    dic["alpha"] = 0.8
    dic["Ta"] = 40.0
    dic["THigh"] = 75.0
    dic["TLow"] = 25.0
    dic["RDCHigh"] = 8.688e-05
    dic["RDCLow"] = 7.283e-05
    dic["azm"] = 90.0
    dic["lat"] = 30.0
    dic["tb"] = 0.0
    dic["alt"] = 0.0
    dic["D"] = 28.14 * 1.0e-03
    dic["d"] = 10.4 * 1.0e-03
    dic["month"] = 6
    dic["day"] = 10
    dic["hour"] = 11.0

    T = 100.0

    assert np.isclose(ieee.ConvectiveCooling(**dic).value(T), 81.93, rtol=0.002)
    assert np.isclose(ieee.RadiativeCooling(**dic).value(T), 39.1, rtol=0.001)
    assert np.isclose(ieee.SolarHeating(**dic).value(T), 22.44, rtol=0.001)
    jh = ieee.JouleHeating(**dic)
    assert np.isclose(jh._rdc(T), 9.390e-05, rtol=1.0e-09)

    # additional debug
    ieee.SolarHeating(**dic).value(T)

    from thermohl import sun

    sd = sun.solar_declination(dic["month"], dic["day"])
    assert np.isclose(np.rad2deg(sd), 23.0, rtol=0.001)

    ha = sun.hour_angle(dic["hour"], minute=0.0, second=0.0)
    assert np.isclose(np.rad2deg(ha), -15.0)

    sa = sun.solar_altitude(
        np.deg2rad(dic["lat"]), dic["month"], dic["day"], dic["hour"]
    )
    assert np.isclose(np.rad2deg(sa), 74.8, rtol=0.002)

    sz = sun.solar_azimuth(
        np.deg2rad(dic["lat"]), dic["month"], dic["day"], dic["hour"]
    )
    np.isclose(np.rad2deg(sz), 114.0, rtol=0.001)

    th = np.arccos(np.cos(sa) * np.cos(sz - dic["azm"]))
    np.isclose(np.rad2deg(th), 76.1, rtol=0.02)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Optional, Any


from thermohl import floatArrayLike, intArrayLike
from thermohl.power import _SRad, SolarHeatingBase


class SolarHeating(SolarHeatingBase):
    def __init__(
        self,
        lat: floatArrayLike,
        alt: floatArrayLike,
        azm: floatArrayLike,
        tb: floatArrayLike,
        month: intArrayLike,
        day: intArrayLike,
        hour: floatArrayLike,
        D: floatArrayLike,
        alpha: floatArrayLike,
        srad: Optional[floatArrayLike] = None,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        lat : float or np.ndarray
            Latitude.
        alt : float or np.ndarray
            Altitude.
        azm : float or np.ndarray
            Azimuth.
        tb : float or np.ndarray
            Air pollution from 0 (clean) to 1 (polluted).
        month : int or np.ndarray
            Month number (must be between 1 and 12).
        day : int or np.ndarray
            Day of the month (must be between 1 and 28, 29, 30 or 31 depending on
            month).
        hour : float or np.ndarray
            Hour of the day (solar, must be between 0 and 23).
        D : float or np.ndarray
            external diameter.
        alpha : float or np.ndarray
            Solar absorption coefficient.
        srad : xxx
            xxx

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        est = _SRad(
            [
                -4.22391e01,
                +6.38044e01,
                -1.9220e00,
                +3.46921e-02,
                -3.61118e-04,
                +1.94318e-06,
                -4.07608e-09,
            ],
            [
                +5.31821e01,
                +1.4211e01,
                +6.6138e-01,
                -3.1658e-02,
                +5.4654e-04,
                -4.3446e-06,
                +1.3236e-08,
            ],
        )
        super().__init__(
            lat, alt, azm, tb, month, day, hour, D, alpha, est, srad, **kwargs
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl import solver

_nprs = 123456


def _solvers(dic=None):
    return [
        solver._factory(dic=dic, heateq="1t", model=m)
        for m in ["rte", "cigre", "ieee", "olla"]
    ]


def test_balance():
    tol = 1.0e-09
    np.random.seed(_nprs)
    N = 9999
    dic = dict(
        lat=np.random.uniform(42.0, 51.0, N),
        alt=np.random.uniform(0.0, 1600.0, N),
        azm=np.random.uniform(0.0, 360.0, N),
        month=np.random.randint(1, 13, N),
        day=np.random.randint(1, 31, N),
        hour=np.random.randint(0, 24, N),
        Ta=np.random.uniform(0.0, 30.0, N),
        ws=np.random.uniform(0.0, 7.0, N),
        wa=np.random.uniform(0.0, 90.0, N),
        I=np.random.uniform(40.0, 4000.0, N),
        d=np.random.randint(2, size=N) * solver.default_values()["d"],
    )

    for s in _solvers(dic):
        df = s.steady_temperature(
            return_err=True, return_power=True, tol=tol, maxiter=64
        )
        assert np.all(df["err"] < tol)
        bl = np.abs(
            df["P_joule"]
            + df["P_solar"]
            - df["P_convection"]
            - df["P_radiation"]
            - df["P_precipitation"]
        )
        atol = np.maximum(
            np.abs(s.balance(df["t"] + 0.5 * df["err"])),
            np.abs(s.balance(df["t"] - 0.5 * df["err"])),
        )
        assert np.allclose(bl, 0.0, atol=atol)


def test_consistency():
    np.random.seed(_nprs)
    N = 9999
    dic = dict(
        lat=np.random.uniform(42.0, 51.0, N),
        alt=np.random.uniform(0.0, 1600.0, N),
        azm=np.random.uniform(0.0, 360.0, N),
        month=np.random.randint(1, 13, N),
        day=np.random.randint(1, 31, N),
        hour=np.random.randint(0, 24, N),
        Ta=np.random.uniform(0.0, 30.0, N),
        ws=np.random.uniform(0.0, 7.0, N),
        wa=np.random.uniform(0.0, 90.0, N),
        d=np.random.randint(2, size=N) * solver.default_values()["d"],
    )

    for s in _solvers(dic):
        df = s.steady_intensity(
            T=100.0, return_err=True, return_power=True, tol=1.0e-09, maxiter=64
        )
        bl = (
            df["P_joule"]
            + df["P_solar"]
            - df["P_convection"]
            - df["P_radiation"]
            - df["P_precipitation"]
        )
        assert np.allclose(bl, 0.0, atol=1.0e-06)
        s.args["I"] = df["I"].values
        s.update()
        dg = s.steady_temperature(
            return_err=True, return_power=True, tol=1.0e-09, maxiter=64
        )
        assert np.allclose(dg["t"].values, 100.0)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

# mypy: ignore-errors
"""Rain cooling power term.

Based on *Modelling precipitation cooling of overhead conductors*, Pytlak et
al., 2011 (https://doi.org/10.1016/j.epsr.2011.06.004) and *Dynamic thermal
rating of power lines in raining conditions-model and measurements*, Maksic et
al., 2016 (https://doi.org/10.1109/PESGM.2016.7741611).
"""

import numpy as np

from thermohl import air
from thermohl import thermodynamics
from thermohl.power import olla, ieee
from thermohl.utils import PowerTerm


class PrecipitationCooling(PowerTerm):
    """Precipitation cooling term."""

    def value(self, T, alt, Ta, ws, wa, D, Pa, rh, pr, **kwargs):
        # maksic = (PrecipitationCooling._evap(T, alt, Ta, ws, wa, D, Pa, rh, pr) +
        #           PrecipitationCooling._imp(T, Ta, ws, D, pr))
        pytlak = PrecipitationCooling._evap(T, alt, Ta, ws, wa, D, Pa, rh, pr)
        return pytlak

    @staticmethod
    def _ma(Ta, ws, D, pr, ps):
        # ! -- pr and ps in m.s**-1
        rho = thermodynamics.Water.volumic_mass(Ta)
        mar = np.sqrt((pr * rho) ** 2 + (ws * 23.589 * pr**0.8460) ** 2)
        mas = np.sqrt((ps * rho) ** 2 + (ws * 142.88 * ps**0.9165) ** 2)
        return D * (mar + mas)

    @staticmethod
    def _me(T, alt, Ta, ws, wa, D, Pa, rh):
        Tf = 0.5 * (T + Ta)
        Td = T - Ta
        vm = air.IEEE.volumic_mass(Tf, alt)
        cc = ieee.ConvectiveCooling._value_forced(Tf, Td, vm, ws, D, wa)
        h = cc / (np.pi * D * Td)
        k = 18.015 / 28.9647
        pm = np.pi * D
        cp = thermodynamics.Air.heat_capacity(T=olla.air.kelvin(Tf))
        ec = thermodynamics.Water.vapor_pressure(T=olla.air.kelvin(T))
        ea = thermodynamics.Water.vapor_pressure(T=olla.air.kelvin(Ta))
        me = pm * h * k * (ec - rh * ea) / (cp * Pa)
        return np.where(Td != 0.0, me, np.zeros_like(T))

    @staticmethod
    def _mass_flux(T, alt, Ta, ws, wa, D, Pa, rh, pr, ps):
        return np.minimum(
            PrecipitationCooling._ma(Ta, ws, D, pr, ps),
            PrecipitationCooling._me(T, alt, Ta, ws, wa, D, Pa, rh),
        )

    @staticmethod
    def _evap(T, alt, Ta, ws, wa, D, Pa, rh, pr, ps=0.0):
        m = PrecipitationCooling._mass_flux(T, alt, Ta, ws, wa, D, Pa, rh, pr, ps)
        Le = thermodynamics.Water.heat_of_vaporization()
        cw = thermodynamics.Water.heat_capacity(T=olla.air.kelvin(T))
        Tb = thermodynamics.Water.boiling_point(p=Pa)
        Tb = air.celsius(Tb)
        Te = np.minimum(T, Tb)
        Lf = thermodynamics.Ice.heat_of_fusion()
        ci = thermodynamics.Ice.heat_capacity()
        rr = m * (Le + cw * (Te - Ta))
        rs = m * (Le + cw * T + Lf * ci * Ta)
        return np.maximum(rr + 0.0 * rs, np.zeros_like(T))

    @staticmethod
    def _imp(T, Ta, ws, D, pr, ps=0.0):
        cw = thermodynamics.Water.heat_capacity(T=olla.air.kelvin(T))
        return 0.71 * cw * (T - Ta) * PrecipitationCooling._ma(Ta, ws, D, pr, ps)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Tuple, Type, Optional, Any, Callable, Final, Dict

import numpy as np

from thermohl import floatArrayLike, floatArray, strListLike, intArray
from thermohl.power import PowerTerm
from thermohl.solver.base import Solver as Solver_, Args
from thermohl.solver.slv3t import Solver3T


class Solver3TL(Solver3T):

    def __init__(
        self,
        dic: Optional[dict[str, Any]] = None,
        joule: Type[PowerTerm] = PowerTerm,
        solar: Type[PowerTerm] = PowerTerm,
        convective: Type[PowerTerm] = PowerTerm,
        radiative: Type[PowerTerm] = PowerTerm,
        precipitation: Type[PowerTerm] = PowerTerm,
    ):
        super().__init__(dic, joule, solar, convective, radiative, precipitation)
        self.update()

    def _morgan_coefficients(self) -> Tuple[floatArray, intArray]:
        """
        Calculate coefficients for heat flux between surface and core in steady state.

        Returns:
        --------
        Tuple[numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[int]]
            - heat_flux_coefficients : numpy.ndarray[float]
                Coefficient array for heat flux.
            - indices_non_zero_diameter : numpy.ndarray[int]
                Indices where core diameter is greater than 0.
                When conductors are uniform, core diameter is equal to 0.0.
                When conductors are bimetallic, core diameter is greater than 0.0.
        """
        UNIFORM_CONDUCTOR_COEFFICIENT: Final[float] = 1 / 13
        BIMETALLIC_CONDUCTOR_COEFFICIENT: Final[float] = 1 / 21

        core_diameter_array = self.args.d * np.ones((self.args.max_len(),))
        indices_non_zero_diameter = np.nonzero(core_diameter_array > 0.0)[0]
        heat_flux_coefficients = UNIFORM_CONDUCTOR_COEFFICIENT * np.ones_like(
            core_diameter_array
        )
        heat_flux_coefficients[indices_non_zero_diameter] = (
            BIMETALLIC_CONDUCTOR_COEFFICIENT
        )
        return heat_flux_coefficients, indices_non_zero_diameter

    def average(self, ts, tc):
        """
        Compute average temperature given surface and core temperature.

        Unlike Solver3T, always use a regular mean even for non-homogeneous
        conductors.

        Parameters:
        ts (numpy.ndarray): Array of surface temperatures.
        tc (numpy.ndarray): Array of core temperatures.

        Returns:
        float or numpy.ndarray: Array of average temperatures.
        """
        return 0.5 * (ts + tc)

    def morgan(self, ts: floatArray, tc: floatArray) -> floatArray:
        """
        Computes the Morgan function for given temperature arrays.

        Parameters:
        ts (numpy.ndarray): Array of surface temperatures.
        tc (numpy.ndarray): Array of core temperatures.

        Returns:
        numpy.ndarray: Resulting array after applying the Morgan function.
        """
        morgan_coefficient = self.mgc[0]
        return (tc - ts) - morgan_coefficient * self.joule(ts, tc)

    def _steady_intensity_header(
        self, T: floatArrayLike, target: strListLike
    ) -> Tuple[np.ndarray, Callable]:
        """Format input for ampacity solver."""

        max_len = self.args.max_len()
        Tmax = T * np.ones(max_len)
        target_ = self._check_target(target, self.args.d, max_len)

        # pre-compute indexes
        surface_indices = np.nonzero(target_ == Solver_.Names.surf)[0]
        average_indices = np.nonzero(target_ == Solver_.Names.avg)[0]
        core_indices = np.nonzero(target_ == Solver_.Names.core)[0]

        def newtheader(i: floatArray, tg: floatArray) -> Tuple[floatArray, floatArray]:
            self.args.I = i
            self.jh.__init__(**self.args.__dict__)
            ts = np.ones_like(tg) * np.nan
            tc = np.ones_like(tg) * np.nan

            ts[surface_indices] = Tmax[surface_indices]
            tc[surface_indices] = tg[surface_indices]

            ts[average_indices] = tg[average_indices]
            tc[average_indices] = 2 * Tmax[average_indices] - ts[average_indices]

            tc[core_indices] = Tmax[core_indices]
            ts[core_indices] = tg[core_indices]

            return ts, tc

        return Tmax, newtheader

    def _morgan_transient(self):
        """Morgan coefficients for transient temperature."""
        c1, _ = self.mgc
        c2 = 0.5 * np.ones_like(c1)
        return c1, c2

    def transient_temperature_legacy(
        self,
        time: floatArray = np.array([]),
        Ts0: Optional[floatArrayLike] = None,
        Tc0: Optional[floatArrayLike] = None,
        tau: float = 600.0,
        return_power: bool = False,
    ) -> Dict[str, Any]:
        """
        Compute transient-state temperature with legacy method.

        Parameters
        ----------
        time : numpy.ndarray
            A 1D array with times (in seconds) when the temperature needs to be
            computed. The array must contain increasing values (undefined
            behaviour otherwise).
        Ts0 : float
            Initial surface temperature. If set to None, the ambient temperature from
            internal dict will be used. The default is None.
        Tc0 : float
            Initial core temperature. If set to None, the ambient temperature from
            internal dict will be used. The default is None.
        tau: float
            A time-constant to add some inertia. The default is 600.
        return_power : bool, optional
            Return power term values. The default is False.

        Returns
        -------
        Dict[str, Any]
            A dictionary with temperature and other results (depending on inputs)
            in the keys.

        """

        # get sizes (n for input dict entries, N for time)
        n = self.args.max_len()
        N = len(time)
        if N < 2:
            raise ValueError()

        # get initial temperature
        Ts0 = Ts0 if Ts0 is not None else self.args.Ta
        Tc0 = Tc0 if Tc0 is not None else 1.0 + Ts0

        # shortcuts for time-loop
        imc = 1.0 / (self.args.m * self.args.c)

        # init
        ts = np.zeros((N, n))
        ta = np.zeros((N, n))
        tc = np.zeros((N, n))
        dT = np.zeros((N, n))

        ts[0, :] = Ts0
        tc[0, :] = Tc0
        ta[0, :] = self.average(ts[0, :], tc[0, :])
        dT[0, :] = tc[0, :] - ts[0, :]

        for i in range(1, len(time)):
            bal = self.balance(ts[i - 1, :], tc[i - 1, :])
            dti = time[i] - time[i - 1]
            ta[i, :] = ta[i - 1, :] + dti * imc * bal
            dT[i, :] = (1.0 - dti / tau) * dT[i - 1, :] + (
                dti / tau * self.mgc[0] * self.jh.value(ta[i, :])
            )
            tc[i, :] = ta[i, :] + 0.5 * dT[i, :]
            ts[i, :] = ta[i, :] - 0.5 * dT[i, :]

        return self._transient_temperature_results(time, ts, ta, tc, return_power, n)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd

from thermohl import solver


def _solvers():
    li = []
    for ht in ["1t", "3t"]:
        for m in ["rte", "cigre", "ieee", "olla"]:
            li.append(solver._factory(dic=None, heateq=ht, model=m))
    return li


def _ampargs(s: solver.Solver, t: pd.DataFrame):
    if isinstance(s, solver.Solver1T):
        a = dict(T=t[solver.Solver.Names.temp].values)
    elif isinstance(s, solver.Solver3T):
        a = dict(T=t[solver.Solver.Names.tsurf].values, target=solver.Solver.Names.surf)
    else:
        raise NotImplementedError
    return a


def _traargs(s: solver.Solver, ds: pd.DataFrame, t, I):
    if isinstance(s, solver.Solver1T):
        a = dict(time=t, T0=ds[solver.Solver.Names.temp].values)
    elif isinstance(s, solver.Solver3T):
        a = dict(
            time=t,
            Ts0=ds[solver.Solver.Names.tsurf].values,
            Tc0=ds[solver.Solver.Names.tcore].values,
        )
    else:
        raise NotImplementedError
    return a


def test_power_default():
    """Check that PowerTerm.value(x) returns correct shape depending on init dict and temperature input."""
    for s in _solvers():
        for p in [s.jh, s.sh, s.cc, s.rc, s.pc]:
            p.__init__(**s.args.__dict__)
            assert np.isscalar(p.value(0.0))
            assert p.value(np.array([0.0])).shape == (1,)
            assert p.value(np.array([0.0, 10.0])).shape == (2,)


def test_power_1d():
    """Check that PowerTerm.value(x) returns correct shape depending on init dict and temperature input."""
    n = 61
    for s in _solvers():
        d = s.args.__dict__.copy()
        d["I"] = np.linspace(0.0, +999.0, n)
        d["alpha"] = np.linspace(0.5, 0.9, n)
        d["Ta"] = np.linspace(-10.0, +50.0, n)
        for p in [s.jh, s.sh, s.cc, s.rc, s.pc]:
            p.__init__(**d)
            v = p.value(0.0)
            assert np.isscalar(v) or v.shape == (n,)
            v = p.value(np.array([0.0]))
            assert v.shape == (1,) or v.shape == (n,)
            assert p.value(np.linspace(-10, +50, n)).shape == (n,)


def test_steady_default():
    for s in _solvers():
        t = s.steady_temperature()
        a = _ampargs(s, t)
        i = s.steady_intensity(**a)
        assert len(t) == 1
        assert len(i) == 1


def test_steady_1d():
    n = 61
    for s in _solvers():
        s.args.Ta = np.linspace(-10, +50, n)
        s.update()
        t = s.steady_temperature()
        a = _ampargs(s, t)
        i = s.steady_intensity(**a)
        assert len(t) == n
        assert len(i) == n


def test_steady_1d_mix():
    n = 61
    for s in _solvers():
        s.args.Ta = np.linspace(-10, +50, n)
        s.args.I = np.array([199.0])
        s.update()
        t = s.steady_temperature()
        a = _ampargs(s, t)
        i = s.steady_intensity(**a)
        assert len(t) == n
        assert len(i) == n


def test_transient_0():
    for s in _solvers():
        t = np.linspace(0, 3600, 361)
        I = 199 * np.ones_like(t)

        ds = s.steady_temperature()
        a = _traargs(s, ds, t, I)

        r = s.transient_temperature(**a)
        assert len(r.pop("time")) == len(t)
        for k in r.keys():
            assert r[k].shape == (len(t),)

        r = s.transient_temperature(**{**a, "return_power": True})

        assert len(r.pop("time")) == len(t)
        for k in r.keys():
            assert r[k].shape == (len(t),)


def test_transient_1():
    n = 7
    for s in _solvers():
        s.args.Ta = np.linspace(-10, +50, n)
        s.update()

        t = np.linspace(0, 3600, 361)
        I = 199 * np.ones_like(t)

        ds = s.steady_temperature()
        a = _traargs(s, ds, t, I)

        r = s.transient_temperature(**a)
        assert len(r.pop("time")) == len(t)
        for k in r.keys():
            assert r[k].shape == (len(t), n)

        r = s.transient_temperature(**{**a, "return_power": True})
        assert len(r.pop("time")) == len(t)
        for k in r.keys():
            assert r[k].shape == (len(t), n)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class JouleHeating(PowerTerm):
    """Joule heating term."""

    @staticmethod
    def _c(
        TLow: floatArrayLike,
        THigh: floatArrayLike,
        RDCLow: floatArrayLike,
        RDCHigh: floatArrayLike,
    ) -> floatArrayLike:
        return (RDCHigh - RDCLow) / (THigh - TLow)

    def __init__(
        self,
        I: floatArrayLike,
        TLow: floatArrayLike,
        THigh: floatArrayLike,
        RDCLow: floatArrayLike,
        RDCHigh: floatArrayLike,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        I : float or np.ndarray
            Transit intensity.
        TLow : float or np.ndarray
            Temperature for RDCHigh measurement.
        THigh : float or np.ndarray
            Temperature for RDCHigh measurement.
        RDCLow : float or np.ndarray
            Electric resistance per unit length at TLow .
        RDCHigh : float or np.ndarray
            Electric resistance per unit length at THigh.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        self.TLow = TLow
        self.THigh = THigh
        self.RDCLow = RDCLow
        self.RDCHigh = RDCHigh
        self.I = I
        self.c = JouleHeating._c(TLow, THigh, RDCLow, RDCHigh)

    def _rdc(self, T: floatArrayLike) -> floatArrayLike:
        return self.RDCLow + self.c * (T - self.TLow)

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        return self._rdc(T) * self.I**2

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating derivative.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        conductor_temperature : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term derivative (W.m\ :sup:`-1`\ K\ :sup:`-1`\ ).

        """
        return self.c * self.I**2 * np.ones_like(conductor_temperature)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class RadiativeCooling(PowerTerm):
    """Power term for radiative cooling.

    Very similar to RadiativeCooling. Difference are in the
    Stefan-Boltzman constant value and the celsius-kelvin conversion.
    """

    def __init__(
        self,
        Ta: floatArrayLike,
        D: floatArrayLike,
        epsilon: floatArrayLike,
        **kwargs: Any,
    ):
        r"""Init with args.

        Parameters
        ----------
        Ta : float or np.ndarray
            Ambient temperature (C).
        D : float or np.ndarray
            External diameter (m).
        epsilon : float or np.ndarray
            Emissivity.

        """
        self.Ta = Ta
        self.D = D
        self.epsilon = epsilon

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute radiative cooling using the Stefan-Boltzmann law.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature (C).

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        return (
            17.8
            * self.epsilon
            * self.D
            * (((T + 273.0) / 100.0) ** 4 - ((self.Ta + 273.0) / 100.0) ** 4)
        )

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Analytical derivative of value method.

        Parameters
        ----------
        conductor_temperature : float or np.ndarray
            Conductor temperature (C).

        Returns
        -------
        float or np.ndarray
            Power term derivative (W.m\ :sup:`-1`\ K\ :sup:`-1`\ ).

        """
        return 4.0 * 1.78e-07 * self.epsilon * self.D * conductor_temperature**3



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class JouleHeating(PowerTerm):
    """Joule heating term."""

    def __init__(
        self,
        I: floatArrayLike,
        km: floatArrayLike,
        kl: floatArrayLike,
        RDC20: floatArrayLike,
        T20: floatArrayLike = 20.0,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        I : float or np.ndarray
            Transit intensity.
        km : float or np.ndarray
            Coefficient for magnetic effects.
        kl : float or np.ndarray
            Linear resistance augmentation with temperature.
        RDC20 : float or np.ndarray
            Electric resistance per unit length (DC) at 20°C.
        T20 : float or np.ndarray, optional
            Reference temperature. The default is 20.

        """
        self.I = I
        self.km = km
        self.kl = kl
        self.RDC20 = RDC20
        self.T20 = T20

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        return self.km * self.RDC20 * (1.0 + self.kl * (T - self.T20)) * self.I**2

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating derivative.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        conductor_temperature : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term derivative (W.m\ :sup:`-1`\ K\ :sup:`-1`\ ).

        """
        return (
            self.km
            * self.RDC20
            * self.kl
            * self.I**2
            * np.ones_like(conductor_temperature)
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl import floatArrayLike


class Air:
    """CIGRE air models."""

    @staticmethod
    def volumic_mass(Tc: floatArrayLike, alt: floatArrayLike = 0.0) -> floatArrayLike:
        r"""Compute air volumic mass.

        If both inputs are numpy arrays, they should have the same size.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius).
        alt : float or numpy.ndarray, optional
            Altitude above sea-level. The default is 0.

        Returns
        -------
        float or numpy.ndarray
             Volumic mass in kg.m\ :sup:`-3`\ .

        """
        return 1.2925 * Air.relative_density(Tc, alt)

    @staticmethod
    def relative_density(
        Tc: floatArrayLike, alt: floatArrayLike = 0.0
    ) -> floatArrayLike:
        """Compute relative density, ie density ratio regarding density at zero altitude.

        This function has temperature and altitude as input for consistency
        regarding other functions in the module, but the temperature has no
        influence, only the altitude for this model.

        If both inputs are numpy arrays, they should have the same size.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius).
        alt : float or numpy.ndarray, optional
            Altitude above sea-level. The default is 0.

        Returns
        -------
        float or numpy.ndarray
            Relative density of air.

        """
        return np.exp(-1.16e-04 * alt) * np.ones_like(Tc)

    @staticmethod
    def kinematic_viscosity(Tc: floatArrayLike) -> floatArrayLike:
        r"""Compute air kinematic viscosity.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)

        Returns
        -------
        float or numpy.ndarray
             Kinematic viscosity in m\ :sup:`2`\ .s\ :sup:`-1`\ .

        """
        return 1.32e-05 + 9.5e-08 * Tc

    @staticmethod
    def dynamic_viscosity(
        Tc: floatArrayLike, alt: floatArrayLike = 0.0
    ) -> floatArrayLike:
        r"""Compute air dynamic viscosity.

        If both inputs are numpy arrays, they should have the same size.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)
        alt : float or numpy.ndarray, optional
            Altitude above sea-level. The default is 0.

        Returns
        -------
        float or numpy.ndarray
             Dynamic viscosity in kg.m\ :sup:`-1`\ .s\ :sup:`-1`\ .

        """
        return Air.kinematic_viscosity(Tc) * Air.volumic_mass(Tc, alt)

    @staticmethod
    def thermal_conductivity(Tc: floatArrayLike) -> floatArrayLike:
        r"""Compute air thermal conductivity.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)

        Returns
        -------
        float or numpy.ndarray
             Thermal conductivity in W.m\ :sup:`-1`\ .K\ :sup:`-1`\ .

        """
        return 2.42e-02 + 7.2e-05 * Tc

    @staticmethod
    def prandtl(Tc: floatArrayLike) -> floatArrayLike:
        """Compute Prandtl number.

        The Prandtl number (Pr) is a dimensionless number, named after the German
        physicist Ludwig Prandtl, defined as the ratio of momentum diffusivity to
        thermal diffusivity.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)

        Returns
        -------
        float or numpy.ndarray
             Prandtl number (no unit)

        """
        return 0.715 - 2.5e-04 * Tc



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power.rte import Air
from thermohl.power.convective_cooling import ConvectiveCoolingBase


class ConvectiveCooling(ConvectiveCoolingBase):
    """Convective cooling term.

    Very similar to IEEE. The differences are in some coefficient values for air
    constants.
    """

    def __init__(
        self,
        alt: floatArrayLike,
        azm: floatArrayLike,
        Ta: floatArrayLike,
        ws: floatArrayLike,
        wa: floatArrayLike,
        D: floatArrayLike,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        alt : float or np.ndarray
            Altitude.
        azm : float or np.ndarray
            Azimuth.
        Ta : float or np.ndarray
            Ambient temperature.
        ws : float or np.ndarray
            Wind speed.
        wa : float or np.ndarray
            Wind angle regarding north.
        D : float or np.ndarray
            External diameter.

        """
        super().__init__(
            alt,
            azm,
            Ta,
            ws,
            wa,
            D,
            Air.volumic_mass,
            Air.dynamic_viscosity,
            Air.thermal_conductivity,
        )

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute convective cooling.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        Tf = 0.5 * (T + self.Ta)
        Td = T - self.Ta
        # very slight difference with air.IEEE.volumic_mass() in coefficient before alt**2
        vm = (1.293 - 1.525e-04 * self.alt + 6.38e-09 * self.alt**2) / (
            1 + 0.00367 * Tf
        )
        return np.maximum(self._value_forced(Tf, Td, vm), self._value_natural(Td, vm))



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

# mypy: ignore-errors
"""Tools to perform Monte Carlo simulations using the thermOHL steady solvers with uncertain input parameters."""

from typing import Union, Tuple

import numpy as np
import pandas as pd
import scipy.stats
from scipy.stats import circmean
from scipy.stats import circstd
from thermohl import distributions, frozen_dist
from thermohl import solver
from thermohl import utils


def default_uncertainties() -> dict:
    """
    Get default parameters for uncertainties.

    Returns
    -------
    dict
        Dictionnary of default distributions and parameters.

    """
    return utils.add_default_uncertainties({}, warning=False)


def cummean(x: np.ndarray) -> np.ndarray:
    """Cumulative mean."""
    return np.cumsum(x) / (1 + np.array(range(len(x))))


def cumstd(x: np.ndarray) -> np.ndarray:
    """Cumulative std."""
    return np.sqrt(cummean(x**2) - cummean(x) ** 2)


def _get_dist(du: dict, mean: float) -> frozen_dist:
    """Get distribution
    -- based on parameters in dict du and with mean mean"""
    mu = mean
    # set std
    sigma = du["std"]
    if du["relative_std"]:
        sigma *= mean
    # if std is 0., return uniform
    if sigma == 0.0:
        return scipy.stats.uniform(mean, mean)

    # select ditribution
    if du["dist"] == "truncnorm":
        a, b = du["min"], du["max"]
        dist = distributions.truncnorm(a, b, mu, sigma)
    elif du["dist"] == "vonmises":
        dist = distributions.vonmises(np.deg2rad(mu), np.deg2rad(sigma))
    elif du["dist"] == "wrapnorm":
        dist = distributions.wrapnorm(np.deg2rad(mu), np.deg2rad(sigma))
    else:
        raise ValueError("Dist keyword not supported")

    return dist


def _generate_samples(
    dc: dict, i: int, du: dict, ns: int, check: bool = False
) -> Union[dict, Tuple[dict, pd.DataFrame]]:
    """
    Generate random samples for all input parameters affected with a probability distribution.
    """

    # sample dict
    ds = {}
    # check dataframe
    if check:
        cl = [
            "key",
            "dist",
            "mean",
            "std",
            "min",
            "max",
            "s_mean",
            "s_std",
            "s_min",
            "s_max",
            "circular",
        ]
        dk = pd.DataFrame(columns=cl, data=np.zeros((len(dc), len(cl))) * np.nan)
        dk.loc[:, "circular"] = False

    # loop on dict
    for j, k in enumerate(dc):
        if k not in du.keys():
            continue
        dist = du[k]["dist"]
        mean = dc[k][i]
        if dist is None or np.isnan(mean):
            sample = mean * np.ones((ns,), dtype=type(mean))
        else:
            sample = _get_dist(du[k], mean).rvs(ns)
            if dist == "vonmises" or dist == "wrapnorm":
                sample = np.rad2deg(sample) % 360.0
        ds[k] = sample

        if check:
            dk.loc[j, "key"] = k
            dk.loc[j, "dist"] = dist
            dk.loc[j, "mean"] = mean
            if "std" in du[k].keys():
                dk.loc[j, "std"] = du[k]["std"]
                if du[k]["relative_std"]:
                    dk.loc[j, "std"] *= mean
            if "min" in du[k].keys():
                dk.loc[j, "min"] = du[k]["min"]
            if "max" in du[k].keys():
                dk.loc[j, "max"] = du[k]["max"]
            if dist in ["vonmises", "wrapnorm"]:
                dk.loc[j, "s_mean"] = circmean(sample, high=360.0, low=0.0)
                dk.loc[j, "s_std"] = circstd(sample, high=360.0, low=0.0)
                dk.loc[j, "circular"] = True
            else:
                dk.loc[j, "s_mean"] = sample.mean()
                dk.loc[j, "s_std"] = sample.std()
            dk.loc[j, "s_min"] = sample.min()
            dk.loc[j, "s_max"] = sample.max()

    if check:
        return ds, dk

    return ds


def _rdict(
    mode: str, target: str, return_surf: bool, return_core: bool, return_avg: bool
) -> dict:
    """Code factorization"""
    if mode == "temperature":
        rdc = dict(return_core=return_core, return_avg=return_avg, return_power=False)
    elif mode == "intensity":
        rdc = dict(
            target=target,
            return_core=return_core,
            return_avg=return_avg,
            return_surf=return_surf,
            return_power=False,
        )
    else:
        raise ValueError("")
    return rdc


def _compute(mode: str, s: solver.Solver, tmx: Union[float, np.ndarray], rdc: dict):
    """Code factorization"""
    if mode == "temperature":
        r = s.steady_temperature(**rdc)
    elif mode == "intensity":
        r = s.steady_intensity(tmx, **rdc)
    else:
        raise ValueError()
    return r


def _steady_uncertainties(
    s: solver.Solver,
    tmax: Union[float, np.ndarray],
    target: str,
    u: dict,
    ns: int,
    return_surf: bool,
    return_core: bool,
    return_avg: bool,
    return_raw: bool,
    mode: str = "temperature",
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, list]]:
    """Code factorization"""
    # return dict
    rdc = _rdict(mode, target, return_surf, return_core, return_avg)

    # save solver dict
    dsave = s.dc

    #  all to max_len size
    n = utils.dict_max_len(s.dc)
    dc = utils.extend_to_max_len(s.dc, n)
    Tmax = tmax * np.ones(
        n,
    )

    # add missing uncertainties parameters
    du = utils.add_default_uncertainties(u)

    # init outputs
    rl = []
    s.dc = solver.default_values()
    tmp = _compute(mode, s, 99.0, rdc)
    cl = []
    for c in tmp.columns:
        cl.append(c + "_mean")
        cl.append(c + "_std")
    dr = pd.DataFrame(data=np.zeros((n, len(cl))), columns=cl)

    # for each entry, generate sample then compute
    for i in range(n):
        s.dc = _generate_samples(dc, i, du, ns, check=False)
        r = _compute(mode, s, Tmax[i], rdc)
        if return_raw:
            rl.append(r)
        mu = r.mean()
        sg = r.std()
        for c in r.columns:
            dr.loc[i, c + "_mean"] = mu[c]
            dr.loc[i, c + "_std"] = sg[c]

    # restore solver dict
    s.dc = dsave

    if return_raw:
        return dr, rl
    else:
        return dr


def temperature(
    s: solver.Solver,
    u: dict = {},
    ns: int = 4999,
    return_core: bool = False,
    return_avg: bool = False,
    return_raw: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, list]]:
    """
    Perform Monte Carlo simulation using the steady temperature solver.
    """
    return _steady_uncertainties(
        s,
        np.nan,
        None,
        u,
        ns,
        None,
        return_core,
        return_avg,
        return_raw,
        mode="temperature",
    )


def intensity(
    s: solver.Solver,
    tmax: Union[float, np.ndarray],
    target: str = "surf",
    u: dict = {},
    ns: int = 4999,
    return_core: bool = False,
    return_avg: bool = False,
    return_surf: bool = False,
    return_raw: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, list]]:
    """
    Perform Monte Carlo simulation using the steady intensity solver.
    """
    return _steady_uncertainties(
        s,
        tmax,
        target,
        u,
        ns,
        return_surf,
        return_core,
        return_avg,
        return_raw,
        mode="intensity",
    )


def _diff_method(
    s: solver.Solver,
    tmax: Union[float, np.ndarray],
    target: str,
    u: dict,
    q: float = 0.95,
    return_surf: bool = False,
    return_core: bool = False,
    return_avg: bool = False,
    ep: float = 1.0e-06,
    mode: str = "temperature",
) -> pd.DataFrame:
    """."""
    # return dict
    rdc = _rdict(
        mode,
        target,
        return_surf=return_surf,
        return_core=return_core,
        return_avg=return_avg,
    )

    # save solver dict
    dsave = s.dc

    #  all to max_len size
    n = utils.dict_max_len(s.dc)
    dc = utils.extend_to_max_len(s.dc, n)

    # add missing uncertainties parameters
    du = utils.add_default_uncertainties(u)

    y0 = _compute(mode, s, tmax, rdc)
    dr = y0 * 0.0
    for k in dc:
        if k not in du.keys() or du[k]["dist"] is None:
            continue
        if np.all(np.isnan(dc[k])):
            continue

        mu = dc[k]
        dx = np.zeros_like(mu)
        for i in range(n):
            if np.isnan(mu[i]):
                dx[i] = 0.0
            else:
                dist = _get_dist(du[k], mu[i])
                try:
                    dx[i] = 0.5 * np.diff(dist.ppf([0.5 * (1 - q), 0.5 * (1 + q)]))[0]
                except ValueError:
                    dx[i] = 0.0
                if np.isnan(dx[i]):
                    dx[i] = 0.0
        s.dc[k] = mu + ep
        yp = _compute(mode, s, tmax, rdc)
        s.dc[k] = mu
        dy = np.abs(yp - y0) / ep
        for c in dy.columns:
            dy.loc[:, c] *= dx
        dr += dy
        if np.any(dy.isna()):
            print("Nans with key %s" % (k,))

    dq = pd.DataFrame()
    for c in y0.columns:
        dq.loc[:, c] = y0.loc[:, c]
        dq.loc[:, c + "_delta"] = dr.loc[:, c]

    # restore solver dict
    s.dc = dsave

    return dq


def temperature_diff(
    s: solver.Solver,
    u: dict,
    q: float = 0.95,
    return_core: bool = False,
    return_avg: bool = False,
    ep: float = 1.0e-06,
) -> pd.DataFrame:
    """."""
    return _diff_method(
        s,
        np.nan,
        None,
        u,
        q=q,
        return_core=return_core,
        return_avg=return_avg,
        ep=ep,
        mode="temperature",
    )


def intensity_diff(
    s: solver.Solver,
    tmax: Union[float, np.ndarray],
    target: str,
    u: dict,
    q: float = 0.95,
    return_core: bool = False,
    return_avg: bool = False,
    return_surf: bool = False,
    ep: float = 1.0e-06,
):
    """."""
    return _diff_method(
        s,
        tmax,
        target,
        u,
        q=q,
        return_core=return_core,
        return_avg=return_avg,
        return_surf=return_surf,
        ep=ep,
        mode="intensity",
    )


def sensitivity(
    s: solver.Solver,
    tmax: Union[float, np.ndarray],
    u: dict,
    ns: int,
    target: str,
    return_surf: bool,
    return_core: bool,
    return_avg: bool,
    mode: str = "temperature",
) -> Tuple[list[pd.DataFrame], list[pd.DataFrame]]:
    """
    Perform a sensitivity analysis with Sobol indices (first order and total indices) using the Monte Carlo method.
    """
    # return dict
    rdc = _rdict(mode, target, return_surf, return_core, return_avg)

    # save solver dict
    dsave = s.dc

    #  all to max_len size
    n = utils.dict_max_len(s.dc)
    dc = utils.extend_to_max_len(s.dc, n)
    Tmax = tmax * np.ones(
        n,
    )

    # add missing uncertainties parameters
    du = utils.add_default_uncertainties(u)

    # init outputs
    d1l = []
    dtl = []

    # for each entry, compute
    for i in range(n):
        # first sample
        smp = _generate_samples(dc, i, du, ns, check=False)
        s.dc = smp
        dp = _compute(mode, s, Tmax[i], rdc)

        # second sample
        smq = _generate_samples(dc, i, du, ns, check=False)
        s.dc = smq
        dq = _compute(mode, s, Tmax[i], rdc)

        pqs = ((dp - dq) ** 2).sum()

        #
        d1 = pd.DataFrame(
            columns=["var"] + dp.columns.tolist(),
            data=np.zeros((len(du), 1 + len(dp.columns))),
        )
        d1.loc[:, "var"] = du.keys()
        dt = pd.DataFrame(
            columns=["var"] + dp.columns.tolist(),
            data=np.zeros((len(du), 1 + len(dp.columns))),
        )
        dt.loc[:, "var"] = du.keys()

        # mix samples, run and compute 1st and total indexes
        for j, k in enumerate(du):
            s.dc = smp.copy()
            s.dc[k] = smq[k]
            dpj = _compute(mode, s, Tmax[i], rdc)

            s.dc = smq.copy()
            s.dc[k] = smp[k]
            dqj = _compute(mode, s, Tmax[i], rdc)

            denom = pqs + ((dpj - dqj) ** 2).sum()
            d1.iloc[j, 1:] = 2.0 * ((dqj - dq) * (dp - dpj)).sum() / denom
            dt.iloc[j, 1:] = ((dq - dqj) ** 2 + (dp - dpj) ** 2).sum() / denom

        d1l.append(d1)
        dtl.append(dt)

    # restore solver dict
    s.dc = dsave

    return d1l, dtl



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Union, List

import numpy as np
import numpy.typing as npt
import scipy

frozen_dist = scipy.stats._distn_infrastructure.rv_continuous_frozen

floatArrayLike = Union[float, npt.NDArray[np.float64]]
intArrayLike = Union[int, npt.NDArray[np.int64]]
numberArrayLike = Union[float, int, npt.NDArray[np.float64], npt.NDArray[np.int64]]
strListLike = Union[str, List[str]]

floatArray = npt.NDArray[np.float64]
intArray = npt.NDArray[np.int64]
numberArray = Union[npt.NDArray[np.float64], npt.NDArray[np.int64]]



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl import floatArrayLike

_zerok = 273.15


def kelvin(t: floatArrayLike) -> floatArrayLike:
    return t + _zerok


class Air:
    """`Wikipedia <https://fr.wikipedia.org/wiki/Air> models."""

    @staticmethod
    def volumic_mass(Tc: floatArrayLike, alt: floatArrayLike = 0.0) -> floatArrayLike:
        r"""
        Compute air volumic mass.

        If both inputs are numpy arrays, they should have the same size.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius).
        alt : float or numpy.ndarray, optional
            Altitude above sea-level. The default is 0.

        Returns
        -------
        float or numpy.ndarray
             Volumic mass in kg.m\ :sup:`-3`\ .

        """
        Tk = kelvin(Tc)
        return 1.292 * _zerok * np.exp(-3.42e-02 * alt / Tk) / Tk

    @staticmethod
    def dynamic_viscosity(Tc: floatArrayLike) -> floatArrayLike:
        r"""Compute air dynamic viscosity.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)

        Returns
        -------
        float or numpy.ndarray
             Dynamic viscosity in kg.m\ :sup:`-1`\ .s\ :sup:`-1`\ .

        """
        Tk = kelvin(Tc)
        return 8.8848e-15 * Tk**3 - 3.2398e-11 * Tk**2 + 6.2657e-08 * Tk + 2.3543e-06

    @staticmethod
    def kinematic_viscosity(
        Tc: floatArrayLike, alt: floatArrayLike = 0.0
    ) -> floatArrayLike:
        r"""Compute air kinematic viscosity.

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)
        alt : float or numpy.ndarray, optional
            Altitude above sea-level. The default is 0.

        Returns
        -------
        float or numpy.ndarray
             Kinematic viscosity in m\ :sup:`2`\ .s\ :sup:`-1`\ .

        """
        return Air.dynamic_viscosity(Tc) / Air.volumic_mass(Tc, alt=alt)

    @staticmethod
    def thermal_conductivity(Tc: floatArrayLike) -> floatArrayLike:
        r"""Compute air thermal conductivity.

        The output is valid for input in [-150, 1300] range (in Celsius)

        Parameters
        ----------
        Tc : float or numpy.ndarray
            Air temperature (in Celsius)

        Returns
        -------
        float or numpy.ndarray
             Thermal conductivity in W.m\ :sup:`-1`\ .K\ :sup:`-1`\ .

        """
        Tk = kelvin(Tc)
        return 1.5207e-11 * Tk**3 - 4.8570e-08 * Tk**2 + 1.0184e-04 * Tk - 3.9333e-04



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Base class to build a solver for heat equation."""

import datetime
from abc import ABC, abstractmethod
from typing import Tuple, Type, Any, Optional, KeysView, Dict

import numpy as np
import pandas as pd
from numpy import ndarray

from thermohl import (
    floatArrayLike,
    floatArray,
    intArray,
    numberArray,
    numberArrayLike,
)
from thermohl.power import PowerTerm


class _DEFPARAM:
    tmin = -99.0
    tmax = +999.0
    tol = 1.0e-09
    maxiter = 64
    imin = 0.0
    imax = 9999.0


class Args:
    """Object to store Solver args in a dict-like manner."""

    # __slots__ = [
    #     'lat', 'lon', 'alt', 'azm', 'month', 'day', 'hour', 'Ta', 'Pa', 'rh', 'pr', 'ws', 'wa', 'al', 'tb', 'I', 'm',
    #     'd', 'D', 'a', 'A', 'R', 'l', 'c', 'alpha', 'epsilon', 'RDC20', 'km', 'ki', 'kl', 'kq', 'RDCHigh', 'RDCLow',
    #     'THigh', 'TLow'
    # ]

    def __init__(self, dic: Optional[dict[str, Any]] = None):
        # add default values
        self._set_default_values()
        # use values from input dict
        if dic is None:
            dic = {}
        keys = self.keys()
        for k in dic:
            if k in keys and dic[k] is not None:
                self[k] = dic[k]

    def _set_default_values(self) -> None:
        """Set default values."""

        self.lat = 45.0  # latitude (deg)
        self.lon = 0.0  # longitude (deg)
        self.alt = 0.0  # altitude (m)
        self.azm = 0.0  # azimuth (deg)

        self.month = 3  # month number (1=Jan, 2=Feb, ...)
        self.day = 21  # day of the month
        self.hour = 12  # hour of the day (in [0, 23] range)

        self.Ta = 15.0  # ambient temperature (C)
        self.Pa = 1.0e05  # ambient pressure (Pa)
        self.rh = 0.8  # relative humidity (none, in [0, 1])
        self.pr = 0.0  # rain precipitation rate (m.s**-1)
        self.ws = 0.0  # wind speed (m.s**-1)
        self.wa = 90.0  # wind angle (deg, regarding north)
        self.al = 0.8  # albedo (1)
        # coefficient for air pollution from 0 (clean) to 1 (polluted)
        self.tb = 0.1

        self.I = 100.0  # transit intensity (A)

        self.m = 1.5  # mass per unit length (kg.m**-1)
        self.d = 1.9e-02  # core diameter (m)
        self.D = 3.0e-02  # external (global) diameter (m)
        self.a = 2.84e-04  # core section (m**2)
        self.A = 7.07e-04  # external (global) section (m**2)
        self.R = 4.0e-02  # roughness (1)
        self.l = 1.0  # radial thermal conductivity (W.m**-1.K**-1)
        self.c = 500.0  # specific heat capacity (J.kg**-1.K**-1)

        self.alpha = 0.5  # solar absorption (1)
        self.epsilon = 0.5  # emissivity (1)
        # electric resistance per unit length (DC) at 20°C (Ohm.m**-1)
        self.RDC20 = 2.5e-05

        self.km = 1.006  # coefficient for magnetic effects (1)
        self.ki = 0.016  # coefficient for magnetic effects (A**-1)
        # linear resistance augmentation with temperature (K**-1)
        self.kl = 3.8e-03
        # quadratic resistance augmentation with temperature (K**-2)
        self.kq = 8.0e-07
        # electric resistance per unit length (DC) at THigh (Ohm.m**-1)
        self.RDCHigh = 3.05e-05
        # electric resistance per unit length (DC) at TLow (Ohm.m**-1)
        self.RDCLow = 2.66e-05
        self.THigh = 60.0  # temperature for RDCHigh measurement (°C)
        self.TLow = 20.0  # temperature for RDCLow measurement (°C)

    def keys(self) -> KeysView[str]:
        """Get list of members as dict keys."""
        return self.__dict__.keys()

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.__dict__[key] = value

    def max_len(self) -> int:
        """
        Calculate the maximum length of the values in the dictionary.

        This method iterates over all keys in the dictionary and determines the maximum length
        of the values associated with those keys.
        If a value is not of a type that has a length, it is ignored.

        Returns:
            int: The maximum length of the values in the dictionary. If the dictionary is empty
            or all values are of types that do not have a length, the method returns 1.
        """
        n = 1
        for k in self.keys():
            try:
                n = max(n, len(self[k]))
            except TypeError:
                pass
        return n

    def extend_to_max_len(self) -> None:
        """
        Extend all elements in the dictionary to the maximum length.

        This method iterates over all keys in the dictionary and checks if the
        corresponding value is a numpy ndarray. If it is, it checks if its length
        matches the maximum length obtained from the `max_len` method.
        If the length matches, it creates a copy of the array.
        If the length does not match or for non-ndarray values, it creates
        a new numpy array of the maximum length, filled with the original value
        and having the same data type.

        Returns:
            None
        """
        n = self.max_len()
        for k in self.keys():
            if isinstance(self[k], np.ndarray):
                t = self[k].dtype
                c = len(self[k]) == n
            else:
                t = type(self[k])
                c = False
            if c:
                self[k] = self[k][:]
            else:
                self[k] = self[k] * np.ones((n,), dtype=t)

    def compress(self) -> None:
        """
        Compresses the values in the dictionary by replacing numpy arrays with a
        single unique value if all elements in the array are the same.

        Returns:
            None
        """
        for k in self.keys():
            if isinstance(self[k], np.ndarray):
                u = np.unique(self[k])
                if len(u) == 1:
                    self[k] = u[0]


class Solver(ABC):
    """Object to solve a temperature problem.

    The temperature of a conductor is driven by four power terms, two heating
    terms (joule and solar heating) and three cooling terms (convective,
    radiative and precipitation cooling). This class is used to solve a
    temperature problem with the heating and cooling terms passed to its
    __init__ function.
    """

    class Names:
        pjle = "P_joule"
        psol = "P_solar"
        pcnv = "P_convection"
        prad = "P_radiation"
        ppre = "P_precipitation"
        err = "err"
        surf = "surf"
        avg = "avg"
        core = "core"
        time = "time"
        transit = "I"
        temp = "t"
        tsurf = "t_surf"
        tavg = "t_avg"
        tcore = "t_core"

        @staticmethod
        def powers() -> tuple[str, str, str, str, str]:
            return (
                Solver.Names.pjle,
                Solver.Names.psol,
                Solver.Names.pcnv,
                Solver.Names.prad,
                Solver.Names.ppre,
            )

    def __init__(
        self,
        dic: Optional[dict[str, Any]] = None,
        joule: Type[PowerTerm] = PowerTerm,
        solar: Type[PowerTerm] = PowerTerm,
        convective: Type[PowerTerm] = PowerTerm,
        radiative: Type[PowerTerm] = PowerTerm,
        precipitation: Type[PowerTerm] = PowerTerm,
    ) -> None:
        """Create a Solver object.

        Parameters
        ----------
        dc : dict
            Input values used in power terms. If there is a missing value, a
            default is used.
        jouleH : utils.PowerTerm
            Joule heating term.
        solarH : utils.PowerTerm
            Solar heating term.
        convectiveC : utils.PowerTerm
            Convective cooling term.
        radiativeC : utils.PowerTerm
            Radiative cooling term.

        Returns
        -------
        None.

        """
        self.args = Args(dic)
        self.args.extend_to_max_len()
        self.jh = joule(**self.args.__dict__)
        self.sh = solar(**self.args.__dict__)
        self.cc = convective(**self.args.__dict__)
        self.rc = radiative(**self.args.__dict__)
        self.pc = precipitation(**self.args.__dict__)
        self.args.compress()

    def update(self) -> None:
        self.args.extend_to_max_len()
        self.jh.__init__(**self.args.__dict__)
        self.sh.__init__(**self.args.__dict__)
        self.cc.__init__(**self.args.__dict__)
        self.rc.__init__(**self.args.__dict__)
        self.pc.__init__(**self.args.__dict__)
        self.args.compress()

    def balance(self, T: floatArrayLike) -> floatArrayLike:
        return (
            self.jh.value(T)
            + self.sh.value(T)
            - self.cc.value(T)
            - self.rc.value(T)
            - self.pc.value(T)
        )

    @abstractmethod
    def steady_temperature(self) -> pd.DataFrame:
        raise NotImplementedError

    @abstractmethod
    def transient_temperature(self) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def steady_intensity(self) -> pd.DataFrame:
        raise NotImplementedError


def reshape(input_array: numberArrayLike, nb_row: int, nb_columns: int) -> numberArray:
    """
    Reshape the input array to the specified dimensions (nr, nc) if possible.

    Args:
        input_array (numberArrayLike): Input array to be reshaped.
        nb_row (int): Desired number of rows for the reshaped array.
        nb_columns (int): Desired number of columns for the reshaped array.

    Returns:
        numberArray: Reshaped array of size (nb_row, nb_columns). If reshaping is not possible,
            returns an array filled with the input_value repeated to fill the dimension (nb_row, nb_columns).

    Raises:
        AttributeError: If the input_array has an invalid shape that cannot be reshaped.
    """
    reshaped_array = ndarray
    try:
        input_shape = input_array.shape
        if len(input_shape) == 1:
            if nb_row == input_shape[0]:
                reshaped_array = np.column_stack(nb_columns * (input_array,))
            elif nb_columns == input_shape[0]:
                reshaped_array = np.vstack(nb_row * (input_array,))
        elif len(input_shape) == 0:
            raise AttributeError()
        else:
            reshaped_array = np.reshape(input_array, (nb_row, nb_columns))
    except AttributeError:
        reshaped_array = input_array * np.ones(
            (nb_row, nb_columns), dtype=type(input_array)
        )
    return reshaped_array


def _set_dates(
    month: floatArrayLike,
    day: floatArrayLike,
    hour: floatArrayLike,
    time: floatArray,
    n: int,
) -> Tuple[intArray, intArray, floatArray]:
    """
    Set months, days and hours as 2D arrays.

    This function is used in transient temperature computations. Inputs month,
    day and hour are floats or 1D arrays of size n; input t is a time vector of
    size N with evaluation times in seconds. It sets arrays months, days and
    hours, of size (N, n) such that
        months[i, j] = datetime(month[j], day[j], hour[j]) + t[i] .

    Args:
        month (floatArrayLike): Array of floats or float representing the months.
        day (floatArrayLike): Array of floats or float representing the days.
        hour (floatArrayLike): Array of floats or float representing the hours.
        time (floatArray): Array of floats representing the time vector in seconds.
        n (int): Size of the input arrays month, day, and hour.

    Returns:
    Tuple[intArray, intArray, floatArray]:
        - months (intArray): 2D array of shape (N, n) with month values.
        - days (intArray): 2D array of shape (N, n) with day values.
        - hours (floatArray): 2D array of shape (N, n) with hour values.
    """
    oi = np.ones((n,), dtype=int)
    of = np.ones((n,), dtype=float)
    month2 = month * oi
    day2 = day * oi
    hour2 = hour * of

    N = len(time)
    months = np.zeros((N, n), dtype=int)
    days = np.zeros((N, n), dtype=int)
    hours = np.zeros((N, n), dtype=float)

    td = np.array(
        [datetime.timedelta()]
        + [
            datetime.timedelta(seconds=float(time[i] - time[i - 1]))
            for i in range(1, N)
        ]
    )

    for j in range(n):
        hj = int(np.floor(hour2[j]))
        dj = datetime.timedelta(seconds=float(3600.0 * (hour2[j] - hj)))
        t0 = datetime.datetime(year=2000, month=month2[j], day=day2[j], hour=hj) + dj
        ts = pd.Series(t0 + td)
        months[:, j] = ts.dt.month
        days[:, j] = ts.dt.day
        hours[:, j] = (
            ts.dt.hour
            + ts.dt.minute / 60.0
            + (ts.dt.second + ts.dt.microsecond * 1.0e-06) / 3600.0
        )

    return months, days, hours



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm
from thermohl.power.cigre import Air


class ConvectiveCooling(PowerTerm):
    """Convective cooling term."""

    def __init__(
        self,
        alt: floatArrayLike,
        azm: floatArrayLike,
        Ta: floatArrayLike,
        ws: floatArrayLike,
        wa: floatArrayLike,
        D: floatArrayLike,
        R: floatArrayLike,
        g: float = 9.81,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        alt : float or np.ndarray
            Altitude.
        azm : float or np.ndarray
            Azimuth.
        Ta : float or np.ndarray
            Ambient temperature.
        ws : float or np.ndarray
            Wind speed.
        wa : float or np.ndarray
            Wind angle regarding north.
        D : float or np.ndarray
            External diameter.
        R : float or np.ndarray
            Cable roughness.
        g : float, optional
            Gravitational acceleration. The default is 9.81.

        """
        self.alt = alt
        self.Ta = Ta
        self.ws = ws
        self.D = D
        self.R = R
        self.g = g
        self.da = np.arcsin(np.sin(np.deg2rad(np.abs(azm - wa) % 180.0)))

    def _nu_forced(self, Tf: floatArrayLike, nu: floatArrayLike) -> floatArrayLike:
        """
        Calculate the Nusselt number for forced convection.

        Parameters:
        -----------
        Tf : float or numpy.ndarray
            Array of fluid temperatures.
        nu : float or numpy.ndarray
            Array of kinematic viscosities.

        Returns:
        --------
        float or numpy.ndarray
            Array of Nusselt numbers for forced convection.
        Notes:
        ------
        The function calculates the Nusselt number based on the relative density of air,
        Reynolds number, and empirical correlations. The correlations are adjusted based
        on the Reynolds number and the ratio of characteristic length to diameter (R).
        The function also considers the angle of attack (da) to adjust the coefficients
        for the Nusselt number calculation.
        """
        relative_density = Air.relative_density(Tf, self.alt)
        Re = relative_density * np.abs(self.ws) * self.D / nu

        s = np.ones_like(Tf) * np.ones_like(nu) * np.ones_like(Re)
        z = s.shape == ()
        if z:
            s = np.array([1.0])

        B1 = 0.641 * s
        n = 0.471 * s

        # NB : (0.641/0.178)**(1/(0.633-0.471)) = 2721.4642715250125
        ix = np.logical_and(self.R <= 0.05, Re >= 2721.4642715250125)
        # NB : (0.641/0.048)**(1/(0.800-0.471)) = 2638.3210085195865
        jx = np.logical_and(self.R > 0.05, Re >= 2638.3210085195865)

        B1[ix] = 0.178
        B1[jx] = 0.048

        n[ix] = 0.633
        n[jx] = 0.800

        if z:
            B1 = B1[0]
            n = n[0]

        B2 = np.where(self.da < np.deg2rad(24.0), 0.68, 0.58)
        m1 = np.where(self.da < np.deg2rad(24.0), 1.08, 0.90)

        return np.maximum(0.42 + B2 * np.sin(self.da) ** m1, 0.55) * (B1 * Re**n)

    def _nu_natural(
        self,
        Tf: floatArrayLike,
        Td: floatArrayLike,
        nu: floatArrayLike,
    ) -> floatArrayLike:
        """
        Calculate the Nusselt number for natural convection.

        Parameters
        ----------
        Tf : float or numpy.ndarray
            Film temperature in degrees Celsius.
        Td : float or numpy.ndarray
            Temperature difference in degrees Celsius.
        nu : float or numpy.ndarray
            Kinematic viscosity in m^2/s.

        Returns
        -------
        float or numpy.ndarray
            Nusselt number for natural convection.

        Notes
        -----
        The function calculates the Grashof number (gr) and the product of the Grashof
        number and the Prandtl number (gp). It then uses these values to determine the
        Nusselt number based on empirical correlations for different ranges of gp.

        """
        gr = self.D**3 * np.abs(Td) * self.g / ((Tf + 273.15) * nu**2)
        gp = gr * Air.prandtl(Tf)
        ia = gp < 1.0e04
        A2 = np.ones_like(gp) * 0.480
        m2 = np.ones_like(gp) * 0.250

        if len(gp.shape) == 0:
            if ia:
                A2 = 0.850
                m2 = 0.188
        else:
            A2[ia] = 0.850
            m2[ia] = 0.188
        return A2 * gp**m2

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute convective cooling.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature.

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        Tf = 0.5 * (T + self.Ta)
        Td = T - self.Ta
        nu = Air.kinematic_viscosity(Tf)
        # nu[nu < 1.0E-06] = 1.0E-06
        lm = Air.thermal_conductivity(Tf)
        # lm[lm < 0.01] = 0.01
        nf = self._nu_forced(Tf, nu)
        nn = self._nu_natural(Tf, Td, nu)
        return np.pi * lm * (T - self.Ta) * np.maximum(nf, nn)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from thermohl.power import ieee


class RadiativeCooling(ieee.RadiativeCooling):
    pass



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from abc import ABC
from typing import Any

import numpy as np

from thermohl import floatArrayLike

_dT = 1.0e-03


class PowerTerm(ABC):
    """Base class for power term."""

    def __init__(self, **kwargs: Any):
        pass

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute power term value in function of temperature.

        Usually this function should be overridden in children classes; if it is
        not the case it will just return zero.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature (C).

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        return np.zeros_like(T) if not np.isscalar(T) else 0.0

    def derivative(
        self, conductor_temperature: floatArrayLike, dT: float = _dT
    ) -> floatArrayLike:
        r"""Compute power term derivative regarding temperature in function of temperature.

        Usually this function should be overriden in children classes; if it is
        not the case it will evaluate the derivative from the value method with
        a second-order approximation.

        Parameters
        ----------
        conductor_temperature : float or np.ndarray
            Conductor temperature (C).
        dT : float, optional
            Temperature increment. The default is 1.0E-03.

        Returns
        -------
        float or np.ndarray
            Power term derivative (W.m\ :sup:`-1`\ K\ :sup:`-1`\ ).

        """
        return (
            self.value(conductor_temperature + dT)
            - self.value(conductor_temperature - dT)
        ) / (2.0 * dT)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Generic radiative cooling term."""

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power.power_term import PowerTerm


class RadiativeCoolingBase(PowerTerm):
    """Generic power term for radiative cooling."""

    def _celsius2kelvin(self, T: floatArrayLike) -> floatArrayLike:
        return T + self.zerok

    def __init__(
        self,
        Ta: floatArrayLike,
        D: floatArrayLike,
        epsilon: floatArrayLike,
        sigma: float = 5.67e-08,
        zerok: float = 273.15,
        **kwargs: Any,
    ):
        r"""Init with args.

        Parameters
        ----------
        Ta : float or np.ndarray
            Ambient temperature (C).
        D : float or np.ndarray
            External diameter (m).
        epsilon : float or np.ndarray
            Emissivity.
        sigma : float, optional
            Stefan-Boltzmann constant in W.m\ :sup:`-2`\ K\ :sup:`4`\ . The
            default is 5.67E-08.
        zerok : float, optional
            Value for zero kelvin.

        Returns
        -------

        """
        self.zerok = zerok
        self.Ta = self._celsius2kelvin(Ta)
        self.D = D
        self.epsilon = epsilon
        self.sigma = sigma

    def value(self, T: floatArrayLike) -> floatArrayLike:
        r"""Compute radiative cooling using the Stefan-Boltzmann law.

        Parameters
        ----------
        T : float or np.ndarray
            Conductor temperature (C).

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        return (
            np.pi
            * self.sigma
            * self.epsilon
            * self.D
            * (self._celsius2kelvin(T) ** 4 - self.Ta**4)
        )

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Analytical derivative of value method.

        Parameters
        ----------
        conductor_temperature : float or np.ndarray
        Conductor temperature (C).

        Returns
        -------
        float or np.ndarray
            Power term derivative (W.m\ :sup:`-1`\ K\ :sup:`-1`\ ).

        """
        return (
            4.0 * np.pi * self.sigma * self.epsilon * self.D * conductor_temperature**3
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Optional, Any

from thermohl.power import ieee
from thermohl import floatArrayLike, intArrayLike


class SolarHeating(ieee.SolarHeating):
    """Solar heating term."""

    def __init__(
        self,
        lat: floatArrayLike,
        alt: floatArrayLike,
        azm: floatArrayLike,
        month: intArrayLike,
        day: intArrayLike,
        hour: floatArrayLike,
        D: floatArrayLike,
        alpha: floatArrayLike,
        srad: Optional[floatArrayLike] = None,
        **kwargs: Any,
    ):
        r"""Init with args.

        See ieee.SolarHeating; it is exactly the same with altitude and
        turbidity set to zero. If more than one input are numpy arrays, they
        should have the same size.

        Parameters
        ----------
        lat : float or np.ndarray
            Latitude.
        alt : float or np.ndarray
             Lltitude.
        azm : float or np.ndarray
            Azimuth.
        month : int or np.ndarray
            Month number (must be between 1 and 12).
        day : int or np.ndarray
            Day of the month (must be between 1 and 28, 29, 30 or 31 depending on
            month).
        hour : float or np.ndarray
            Hour of the day (solar, must be between 0 and 23).
        D : float or np.ndarray
            external diameter.
        alpha : float or np.ndarray
            Solar absorption coefficient.
        srad : xxx
            xxx

        Returns
        -------
        float or np.ndarray
            Power term value (W.m\ :sup:`-1`\ ).

        """
        if "tb" in kwargs.keys():
            kwargs.pop("tb")
        super().__init__(
            lat=lat,
            alt=alt,
            azm=azm,
            tb=0.0,
            month=month,
            day=day,
            hour=hour,
            D=D,
            alpha=alpha,
            srad=srad,
            **kwargs,
        )
